from ..grammar import McCompParser as Parser, McCompVisitor
from .comp import Comp
from ..common import ComponentParameter, Expr, MetaData
from zenlog import log

class CompVisitor(McCompVisitor):
    def __init__(self, parent, filename, instance_name=None):
        self.parent = parent  # the instrument (handler?) that wanted to read this component
        self.filename = filename
        self.state = Comp()
        self.instance_name = instance_name

    def visitProg(self, ctx: Parser.ProgContext):
        self.state = Comp()
        self.visit(ctx.component_definition())
        return self.state

    def visitComponentDefineNew(self, ctx: Parser.ComponentDefineNewContext):
        self.state.name = str(ctx.Identifier())
        if ctx.NoAcc() is not None:
            self.state.no_acc()
        self.visitChildren(ctx)  # Use the visitor methods to fill in details of the state

    def visitComponentDefineCopy(self, ctx: Parser.ComponentDefineCopyContext):
        from copy import deepcopy
        new_name, copy_from = [str(x) for x in ctx.Identifier()]
        if self.parent is None:
            raise RuntimeError("Can not copy a component definition without a parent instrument")
        copy_from_comp = self.parent.get_component(copy_from)
        # pull from that component ... deepcopy _just_ in case
        self.state = deepcopy(copy_from_comp)
        # update the name to match what we're going to call this component
        self.state.name = new_name
        # add parameters, overwrite all provided details
        if ctx.NoAcc() is not None:
            self.state.no_acc()
        self.visitChildren(ctx)  # Use the visitor methods to overwrite details of the state

    def visitCategory(self, ctx: Parser.CategoryContext):
        # Return the provided identifier or string literal, minus quotes
        self.state.category = str(ctx.StringLiteral())[1:-1] if ctx.Identifier() is None else str(ctx.Identifer())

    def visitTraceBlock(self, ctx: Parser.TraceBlockContext):
        self.state.TRACE(self.visit(ctx.unparsed_block()))
        
    def visitTraceBlockCopy(self, ctx: Parser.TraceBlockCopyContext):
        if len(self.state.trace):
            log.critical(f'The component {self.state.name} is a copied definition, complete with TRACE section')
            log.critical(f'Now `TRACE COPY {ctx.Identifier()}` would add to the existing copied TRACE')
            log.critical(f'This is almost certainly not the intended behaviour, so you get only the new copy.')
            self.state.trace = ()
        blocks = self.parent.get_component(str(ctx.Identifier())).trace
        if ctx.Extend() is not None:
            blocks = [x for x in blocks]
            blocks.append(self.visit(ctx.unparsed_block()))
        self.state.TRACE(*blocks)

    def visitComponent_define_parameters(self, ctx: Parser.Component_define_parametersContext):
        for parameter in self.visit(ctx.component_parameters()):
            self.state.add_parameter(parameter)

    def visitComponent_set_parameters(self, ctx: Parser.Component_set_parametersContext):
        for parameter in self.visit(ctx.component_parameters()):
            self.state.add_setting(parameter)

    def visitComponent_out_parameters(self, ctx: Parser.Component_out_parametersContext):
        for parameter in self.visit(ctx.component_parameters()):
            self.state.add_output(parameter)

    def visitComponent_parameters(self, ctx: Parser.Component_parametersContext):
        return [self.visit(x) for x in ctx.component_parameter()]

    def visitComponentParameterDouble(self, ctx: Parser.ComponentParameterDoubleContext):
        name = str(ctx.Identifier())
        value = None
        if ctx.Assign() is not None:
            # protect against a literal '0' provided ... which doesn't match IntegerLiteral for some reason
            value = 0 if ctx.expr() is None else self.visit(ctx.expr())
        return ComponentParameter(name=name, value=Expr.float(value))

    def visitComponentParameterInteger(self, ctx: Parser.ComponentParameterIntegerContext):
        name = str(ctx.Identifier())
        value = None
        if ctx.Assign() is not None:
            # protect against a literal '0' provided ... which doesn't match IntegerLiteral for some reason
            value = 0 if ctx.expr() is None else self.visit(ctx.expr())
        return ComponentParameter(name=name, value=Expr.int(value))

    def visitComponentParameterString(self, ctx: Parser.ComponentParameterStringContext):
        name = str(ctx.Identifier())
        default = None
        if ctx.Assign() is not None:
            default = 'NULL' if ctx.StringLiteral() is None else str(ctx.StringLiteral())
        return ComponentParameter(name=name, value=Expr.str(default))

    def visitComponentParameterVector(self, ctx: Parser.ComponentParameterVectorContext):
        from ..common import Value, DataType, ShapeType
        name = str(ctx.Identifier(0))
        if ctx.Assign() is not None and ctx.initializerlist() is not None:
            value = self.visit(ctx.initializerlist())
            value.data_type = DataType.float
        else:
            default = None
            if ctx.Assign() is not None:
                default = "NULL"
                if ctx.Identifier(1) is not None:
                    default = str(ctx.Identifier(1))
            value = Expr(Value(default, data_type=DataType.float, shape_type=ShapeType.vector))
        return ComponentParameter(name=name, value=value)

    def visitComponentParameterSymbol(self, ctx: Parser.ComponentParameterSymbolContext):
        raise RuntimeError("McCode symbol parameter type not supported yet")

    def visitComponentParameterDoubleArray(self, ctx: Parser.ComponentParameterDoubleArrayContext):
        # 'vector' is really just an alias for 'double *', right?
        return self.visitComponentParameterVector(ctx)

    def visitComponentParameterIntegerArray(self, ctx: Parser.ComponentParameterIntegerArrayContext):
        from ..common import Value, DataType, ShapeType
        name = str(ctx.Identifier(0))
        if ctx.assign() is not None and ctx.initializerlist() is not None:
            value = self.visit(ctx.initializerlist())
            value.data_type = DataType.int
        else:
            default = None
            if ctx.assign() is not None:
                default = "NULL"
                if ctx.Identifier(1) is not None:
                    default = str(ctx.Identifier(1))
            value = Expr(Value(default, data_type=DataType.int, shape_type=ShapeType.vector))
        return ComponentParameter(name=name, value=value)

    def visitDependency(self, ctx: Parser.DependencyContext):
        if ctx.StringLiteral() is not None:
            # the flags are the literal string without its quotes:
            self.parent.add_c_flags(str(ctx.StringLiteral()).strip('"'))

    def visitDeclareBlock(self, ctx: Parser.DeclareBlockContext):
        self.state.DECLARE(self.visit(ctx.unparsed_block()))

    def visitDeclareBlockCopy(self, ctx: Parser.DeclareBlockCopyContext):
        if len(self.state.declare):
            log.critical(f'The component {self.state.name} is a copied definition, complete with DECLARE section')
            log.critical(f'Now `DECLARE COPY {ctx.Identifier()}` would add to the existing copied DECLARE')
            log.critical(f'This is almost certainly not the intended behaviour, so you get only the new copy.')
            self.state.declare = ()
        blocks = self.parent.get_component(str(ctx.Identifier())).declare
        if ctx.Extend() is not None:
            blocks = [x for x in blocks]
            blocks.append(self.visit(ctx.unparsed_block()))
        self.state.DECLARE(*blocks)

    def visitShareBlock(self, ctx: Parser.ShareBlockContext):
        self.state.SHARE(self.visit(ctx.unparsed_block()))

    def visitShareBlockCopy(self, ctx: Parser.ShareBlockCopyContext):
        blocks = self.parent.get_component(str(ctx.Identifier())).share
        if ctx.Extend() is not None:
            blocks = [x for x in blocks]
            blocks.append(self.visit(ctx.unparsed_block()))
        self.state.SHARE(*blocks)

    def visitInitializeBlock(self, ctx: Parser.InitializeBlockContext):
        self.state.INITIALIZE(self.visit(ctx.unparsed_block()))

    def visitInitializeBlockCopy(self, ctx: Parser.InitializeBlockCopyContext):
        if len(self.state.initialize):
            log.critical(f'The component {self.state.name} is a copied definition, complete with INITIALIZE section')
            log.critical(f'Now `INITIALIZE COPY {ctx.Identifier()}` would add to the existing copied INITIALIZE')
            log.critical(f'This is almost certainly not the intended behaviour, so you get only the new copy.')
            self.state.initialize = ()
        blocks = self.parent.get_component(str(ctx.Identifier())).initialize
        if ctx.Extend() is not None:
            blocks = [x for x in blocks]
            blocks.append(self.visit(ctx.unparsed_block()))
        self.state.INITIALIZE(*blocks)

    def visitUservars(self, ctx: Parser.UservarsContext):
        self.state.USERVARS(self.visit(ctx.unparsed_block()))

    def visitSaveBlock(self, ctx: Parser.SaveBlockContext):
        self.state.SAVE(self.visit(ctx.unparsed_block()))

    def visitSaveBlockCopy(self, ctx: Parser.SaveBlockCopyContext):
        if len(self.state.save):
            log.critical(f'The component {self.state.name} is a copied definition, complete with SAVE section')
            log.critical(f'Now `SAVE COPY {ctx.Identifier()}` would add to the existing copied SAVE')
            log.critical(f'This is almost certainly not the intended behaviour, so you get only the new copy.')
            self.state.save = ()
        blocks = self.parent.get_component(str(ctx.Identifier())).save
        if ctx.Extend() is not None:
            blocks = [x for x in blocks]
            blocks.append(self.visit(ctx.unparsed_block()))
        self.state.SAVE(*blocks)

    def visitFinallyBlock(self, ctx: Parser.FinallyBlockContext):
        self.state.FINALLY(self.visit(ctx.unparsed_block()))

    def visitFinallyBlockCopy(self, ctx: Parser.FinallyBlockCopyContext):
        if len(self.state.final):
            log.critical(f'The component {self.state.name} is a copied definition, complete with FINALLY section')
            log.critical(f'Now `FINALLY COPY {ctx.Identifier()}` would add to the existing copied FINALLY')
            log.critical(f'This is almost certainly not the intended behaviour, so you get only the new copy.')
            self.state.final = ()
        blocks = self.parent.get_component(str(ctx.Identifier())).final
        if ctx.Extend() is not None:
            blocks = [x for x in blocks]
            blocks.append(self.visit(ctx.unparsed_block()))
        self.state.FINALLY(*blocks)

    def visitDisplayBlock(self, ctx: Parser.DisplayBlockContext):
        self.state.DISPLAY(self.visit(ctx.unparsed_block()))

    def visitDisplayBlockCopy(self, ctx: Parser.DisplayBlockCopyContext):
        if len(self.state.display):
            log.critical(f'The component {self.state.name} is a copied definition, complete with DISPLAY section')
            log.critical(f'Now `DISPLAY COPY {ctx.Identifier()}` would add to the existing copied DISPLAY')
            log.critical(f'This is almost certainly not the intended behaviour, so you get only the new copy.')
            self.state.display = ()
        blocks = self.parent.get_component(str(ctx.Identifier())).display
        if ctx.Extend() is not None:
            blocks = [x for x in blocks]
            blocks.append(self.visit(ctx.unparsed_block()))
        self.state.DISPLAY(*blocks)

    def visitMetadata(self, ctx: Parser.MetadataContext):
        filename, line_number, metadata = self.visit(ctx.unparsed_block())
        mime = ctx.mime.text if ctx.mime.type == Parser.Identifier else ctx.mime.text[1:-1]
        name = ctx.name.text if ctx.name.type == Parser.Identifier else ctx.name.text[1:-1]
        metadata = MetaData.from_component_tokens(source=self.state.name, mimetype=mime, name=name, value=metadata)
        self.state.add_metadata(metadata)

    def visitUnparsed_block(self, ctx: Parser.Unparsed_blockContext):
        # We want to extract the source-file line number (and filename) for use in the C-preprocessor
        # via `#line {number} "{filename}"` directives, for more expressive error handling
        line_number = None if ctx.start is None else ctx.start.line
        content = str(ctx.UnparsedBlock())[2:-2]
        return self.filename, line_number, content

    # TODO Make this and the identical list of visitors in instr/visitor.py a single definition ... somehow
    # FIXME There *are* no statements in McCode, so all identifiers always produce un-parsable values.
    def visitAssignment(self, ctx: Parser.AssignmentContext):
        line_number = None if ctx.start is None else ctx.start.line
        raise RuntimeError(f"{self.filename}: {line_number} -- assignment statements are not (yet) supported")

    # TODO (maybe) Add control and statements into McCode, requiring some form of global stack.
    def visitExpressionUnaryPM(self, ctx: Parser.ExpressionUnaryPMContext):
        right = self.visit(ctx.expr())
        if isinstance(right, str):
            return '-' + right if ctx.Plus() is None else right
        return -right if ctx.Plus() is None else right

    def visitExpressionGrouping(self, ctx: Parser.ExpressionGroupingContext):
        from ..common import UnaryOp
        return Expr(UnaryOp('__group__', self.visit(ctx.expr())))

    def visitExpressionFloat(self, ctx: Parser.ExpressionFloatContext):
        return Expr.float(str(ctx.FloatingLiteral()))

    def visitExpressionPointerAccess(self, ctx: Parser.ExpressionPointerAccessContext):
        from ..common import BinaryOp, Value, ObjectType
        pointer = Expr(Value(str(ctx.Identifier()), object_type=ObjectType.identifier))
        return Expr(BinaryOp('__pointer_access__', pointer, self.visit(ctx.expr())))

    def visitExpressionStructAccess(self, ctx: Parser.ExpressionStructAccessContext):
        from ..common import BinaryOp, Value, ObjectType
        struct = Expr(Value(str(ctx.Identifier()), object_type=ObjectType.identifier))
        return Expr(BinaryOp('__struct_access__', struct, self.visit(ctx.expr())))

    def visitExpressionArrayAccess(self, ctx: Parser.ExpressionArrayAccessContext):
        from ..common import BinaryOp, Value, ShapeType, ObjectType
        array = Expr(Value(str(ctx.Identifier()), object_type=ObjectType.identifier, shape_type=ShapeType.vector))
        return Expr(BinaryOp('__getitem__', array, self.visit(ctx.expr())))

    def visitExpressionIdentifier(self, ctx: Parser.ExpressionIdentifierContext):
        from ..common import Value, ObjectType
        return Expr(Value(str(ctx.Identifier()), object_type=ObjectType.identifier))

    def visitExpressionInteger(self, ctx: Parser.ExpressionIntegerContext):
        return Expr.int(str(ctx.IntegerLiteral()))

    def visitExpressionZero(self, ctx: Parser.ExpressionZeroContext):
        return Expr.int(0)

    def visitExpressionExponentiation(self, ctx: Parser.ExpressionExponentiationContext):
        base = self.visit(ctx.base)
        exponent = self.visit(ctx.exponent)
        return base ** exponent

    def visitExpressionBinaryPM(self, ctx: Parser.ExpressionBinaryPMContext):
        left = self.visit(ctx.left)
        right = self.visit(ctx.right)
        return left + right if ctx.Minus() is None else left - right

    def visitExpressionFunctionCall(self, ctx: Parser.ExpressionFunctionCallContext):
        from ..common import BinaryOp, Value, ObjectType
        function = Value(str(ctx.Identifier()), object_type=ObjectType.function)
        args = [self.visit(arg).expr[0] for arg in ctx.args]  # each is a Value, UnaryOp, or BinaryOp?
        return Expr(BinaryOp('__call__', function, args))

    def visitExpressionBinaryMD(self, ctx: Parser.ExpressionBinaryMDContext):
        left, right = self.visit(ctx.left), self.visit(ctx.right)
        return left * right if ctx.Div() is None else left / right

    def visitInitializerlist(self, ctx: Parser.InitializerlistContext):
        from ..common import Value, ObjectType, ShapeType
        values = [self.visit(x).expr[0].value for x in ctx.values]
        return Expr(Value(values, object_type=ObjectType.initializer_list, shape_type=ShapeType.vector))

    def visitExpressionUnaryLogic(self, ctx: Parser.ExpressionUnaryLogicContext):
        from ..common import UnaryOp
        expr = self.visit(ctx.expr())
        op = 'unknown'
        if ctx.Not() is not None:
            op = '__not__'
        return Expr(UnaryOp(op, expr))

    def visitExpressionTrinaryLogic(self, ctx: Parser.ExpressionTrinaryLogicContext):
        from ..common import TrinaryOp
        test, true, false = [self.visit(x) for x in (ctx.test, ctx.true, ctx.false)]
        return Expr(TrinaryOp('__trinary__', test, true, false))

    def visitExpressionBinaryLogic(self, ctx: Parser.ExpressionBinaryLogicContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        op = 'unknown'
        if ctx.AndAnd() is not None:
            op = '__and__'
        elif ctx.OrOr() is not None:
            op = '__or__'
        return Expr(BinaryOp(op, left, right))

    def visitExpressionBinaryEqual(self, ctx: Parser.ExpressionBinaryEqualContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__eq__', left, right))

    def visitExpressionBinaryLessEqual(self, ctx: Parser.ExpressionBinaryLessEqualContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__le__', left, right))

    def visitExpressionBinaryGreaterEqual(self, ctx: Parser.ExpressionBinaryGreaterEqualContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__ge__', left, right))

    def visitExpressionBinaryLess(self, ctx: Parser.ExpressionBinaryLessContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__lt__', left, right))

    def visitExpressionBinaryGreater(self, ctx: Parser.ExpressionBinaryGreaterContext):
        from ..common import BinaryOp
        left, right = [self.visit(x) for x in (ctx.left, ctx.right)]
        return Expr(BinaryOp('__gt__', left, right))

    def visitExpressionString(self, ctx: Parser.ExpressionStringContext):
        strings = ''.join(str(sl).strip('"') for sl in ctx.StringLiteral())
        return Expr.str(f'"{strings}"')

    def visitExpressionPrevious(self, ctx: Parser.ExpressionPreviousContext):
        # The very-special no-good expression use of PREVIOUS where it is replaced by the last component's name
        raise RuntimeError('PREVIOUS is not a valid expression in Comp definitions')

    def visitExpressionMyself(self, ctx: Parser.ExpressionMyselfContext):
        # The even-worse expression use of MYSELF to refer to the current being-constructed component's name
        return Expr.str(self.instance.name)



from .builder import _ensure_antlr_files_up_to_date_on_import


def _import_component_language():
    from .McCompLexer import McCompLexer
    from .McCompParser import McCompParser
    from .McCompListener import McCompListener
    from .McCompVisitor import McCompVisitor
    return McCompLexer, McCompParser, McCompListener, McCompVisitor


def _import_instrument_language():
    from .McInstrLexer import McInstrLexer
    from .McInstrParser import McInstrParser
    from .McInstrListener import McInstrListener
    from .McInstrVisitor import McInstrVisitor
    return McInstrLexer, McInstrParser, McInstrListener, McInstrVisitor


def _import_c_language():
    from .CLexer import CLexer
    from .CParser import CParser
    from .CListener import CListener
    from .CVisitor import CVisitor
    return CLexer, CParser, CListener, CVisitor


# Run the language file check and (re)build the files if necessary
_ensure_antlr_files_up_to_date_on_import('McComp', deps=('McCommon', 'cpp'))
_ensure_antlr_files_up_to_date_on_import('McInstr', deps=('McCommon', 'cpp'))
_ensure_antlr_files_up_to_date_on_import('C')

# Import the classes defined in the language files
McCompLexer, McCompParser, McCompListener, McCompVisitor = _import_component_language()
McInstrLexer, McInstrParser, McInstrListener, McInstrVisitor = _import_instrument_language()
CLexer, CParser, CListener, CVisitor = _import_c_language()

# And set only their names to be exported:
__all__ = [
    'McCompLexer',
    'McCompParser',
    'McCompListener',
    'McCompVisitor',
    'McInstrLexer',
    'McInstrParser',
    'McInstrListener',
    'McInstrVisitor',
    'CLexer',
    'CParser',
    'CListener',
    'CVisitor',
]



from dataclasses import dataclass, field
from ..common import Expr, unary_expr, binary_expr
from zenlog import log
from typing import TypeVar, NamedTuple, Union

VectorType = TypeVar('VectorType', bound='Vector')
AnglesType = TypeVar('AnglesType', bound='Angles')
MatrixType = TypeVar('MatrixType', bound='Matrix')
RotationType = TypeVar('RotationType', bound='Rotation')
SeitzType = TypeVar('SeitzType', bound='Seitz')


class Matrix(NamedTuple):
    """Any 3D matrix, not necessarily a rotation matrix"""
    xx: Expr = Expr.float(0)
    xy: Expr = Expr.float(0)
    xz: Expr = Expr.float(0)
    yx: Expr = Expr.float(0)
    yy: Expr = Expr.float(0)
    yz: Expr = Expr.float(0)
    zx: Expr = Expr.float(0)
    zy: Expr = Expr.float(0)
    zz: Expr = Expr.float(0)

    @classmethod
    def eye(cls):
        o = Expr.float(1)
        z = Expr.float(0)
        return cls(o, z, z, z, o, z, z, z, o)

    def __str__(self):
        x = [f'{x}' for x in (self.xx, self.xy, self.xz, self.yx, self.yy, self.yz, self.zx, self.zy, self.zz)]
        widths = [max(len(x[i+3*j]) for j in range(3)) for i in range(3)]
        line_fmt = [f'>{w+2:d}s' for w in widths]
        lines = '\n'.join(' '.join(f'{x[i+3*j]:{f:s}}' for i, f in zip(range(3), line_fmt)) for j in range(3))
        return lines

    def __repr__(self):
        return str(self)

    def __mul__(self, other: Union[MatrixType, VectorType, Expr]) -> Union[MatrixType, VectorType]:
        if isinstance(other, Matrix):
            return Matrix(
                self.xx * other.xx + self.xy * other.yx + self.xz * other.zx,
                self.xx * other.xy + self.xy * other.yy + self.xz * other.zy,
                self.xx * other.xz + self.xy * other.yz + self.xz * other.zz,
                self.yx * other.xx + self.yy * other.yx + self.yz * other.zx,
                self.yx * other.xy + self.yy * other.yy + self.yz * other.zy,
                self.yx * other.xz + self.yy * other.yz + self.yz * other.zz,
                self.zx * other.xx + self.zy * other.yx + self.zz * other.zx,
                self.zx * other.xy + self.zy * other.yy + self.zz * other.zy,
                self.zx * other.xz + self.zy * other.yz + self.zz * other.zz,
            )
        elif isinstance(other, Vector):
            return Vector(self.xx * other.x + self.xy * other.y + self.xz * other.z,
                          self.yx * other.x + self.yy * other.y + self.yz * other.z,
                          self.zx * other.x + self.zy * other.y + self.zz * other.z)
        else:
            return Matrix(self.xx * other, self.xy * other, self.xz * other,
                          self.yx * other, self.yy * other, self.yz * other,
                          self.zx * other, self.zy * other, self.zz * other)

    def __add__(self, other: MatrixType) -> MatrixType:
        if isinstance(other, Matrix):
            return Matrix(self.xx + other.xx, self.xy + other.xy, self.xz + other.xz,
                          self.yx + other.yx, self.yy + other.yy, self.yz + other.yz,
                          self.zx + other.zx, self.zy + other.zy, self.zz + other.zz)
        else:
            raise RuntimeError(f'No addition with {type(other)}')

    def __sub__(self, other: MatrixType) -> MatrixType:
        if isinstance(other, Matrix):
            return Matrix(self.xx - other.xx, self.xy - other.xy, self.xz - other.xz,
                          self.yx - other.yx, self.yy - other.yy, self.yz - other.yz,
                          self.zx - other.zx, self.zy - other.zy, self.zz - other.zz)
        else:
            raise RuntimeError(f'No subtraction with {type(other)}')

    def __abs__(self) -> MatrixType:
        return Matrix(abs(self.xx), abs(self.xy), abs(self.xz),
                      abs(self.yx), abs(self.yy), abs(self.yz),
                      abs(self.zx), abs(self.zy), abs(self.zz))

    def __round__(self, n=None) -> MatrixType:
        return Matrix(round(self.xx, n), round(self.xy, n), round(self.xz, n),
                      round(self.yx, n), round(self.yy, n), round(self.yz, n),
                      round(self.zx, n), round(self.zy, n), round(self.zz, n))


class Vector(NamedTuple):
    x: Expr = Expr.float(0)
    y: Expr = Expr.float(0)
    z: Expr = Expr.float(0)

    def __mul__(self, other: Union[MatrixType, RotationType, Expr]) -> VectorType:
        if isinstance(other, Expr):
            return Vector(self.x * other, self.y * other, self.z * other)
        return Vector(self.x * other.xx + self.y * other.yx + self.z * other.zx,
                      self.x * other.xy + self.y * other.yy + self.z * other.zy,
                      self.x * other.xz + self.y * other.yz + self.z * other.zz)

    def __truediv__(self, other: Expr) -> VectorType:
        return Vector(self.x / other, self.y / other, self.z / other)

    def __add__(self, other: VectorType) -> VectorType:
        return Vector(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other: VectorType) -> VectorType:
        return Vector(self.x - other.x, self.y - other.y, self.z - other.z)

    def cross_matrix(self) -> Matrix:
        z = Expr.float(0)
        return Matrix(z, -self.z, self.y, self.z, z, -self.x, -self.y, self.x, z)

    def length(self) -> Expr:
        from mccode_antlr.common.expression import unary_expr
        from math import sqrt
        if self.x.is_zero and self.y.is_zero:
            return abs(self.z)
        if self.y.is_zero and self.z.is_zero:
            return abs(self.x)
        if self.z.is_zero and self.x.is_zero:
            return abs(self.y)
        return unary_expr(sqrt, 'sqrt', self.x * self.x + self.y * self.y + self.z * self.z)

    def is_null(self) -> bool:
        return self.x.is_zero and self.y.is_zero and self.z.is_zero

    def __str__(self):
        return f'({self.x}, {self.y}, {self.z})'

    def __contains__(self, value):
        return any(value in x for x in (self.x, self.y, self.z))


class Angles(NamedTuple):
    x: Expr = Expr.float(0)
    y: Expr = Expr.float(0)
    z: Expr = Expr.float(0)

    def __mul__(self, other: Expr) -> AnglesType:
        return Angles(self.x * other, self.y * other, self.z * other)

    def __add__(self, other: AnglesType) -> AnglesType:
        return Angles(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other: AnglesType) -> AnglesType:
        return Angles(self.x - other.x, self.y - other.y, self.z - other.z)

    def is_null(self) -> bool:
        return self.x.is_zero and self.y.is_zero and self.z.is_zero

    def __str__(self):
        return f'({self.x}, {self.y}, {self.z})'

    def __contains__(self, value):
        return any(value in x for x in (self.x, self.y, self.z))


class Rotation(NamedTuple):
    """A _valid_ 3-D rotation matrix flattened in row-order.

    Note:
        To be valid, the Rotation must have an inverse which is its (conjugate) transpose.
        This is equivalent to the matrix being Orthogonal (or Unitary, if complex-valued)
    """
    xx: Expr = Expr.float(1)
    xy: Expr = Expr.float(0)
    xz: Expr = Expr.float(0)
    yx: Expr = Expr.float(0)
    yy: Expr = Expr.float(1)
    yz: Expr = Expr.float(0)
    zx: Expr = Expr.float(0)
    zy: Expr = Expr.float(0)
    zz: Expr = Expr.float(1)

    def __str__(self):
        return f'[({self.xx}, {self.xy}, {self.xz})({self.yx}, {self.yy}, {self.yz})({self.zx}, {self.zy}, {self.zz})]'

    def transpose(self) -> RotationType:
        return Rotation(self.xx, self.yx, self.zx, self.xy, self.yy, self.zy, self.xz, self.yz, self.zz)

    def inverse(self) -> RotationType:
        return self.transpose()

    def __mul__(self, other: Union[RotationType, Vector]) -> Union[RotationType, Vector]:
        if isinstance(other, Vector):
            return Vector(self.xx * other.x + self.xy * other.y + self.xz * other.z,
                          self.yx * other.x + self.yy * other.y + self.yz * other.z,
                          self.zx * other.x + self.zy * other.y + self.zz * other.z)
        elif isinstance(other, Rotation):
            return Rotation(
                self.xx * other.xx + self.xy * other.yx + self.xz * other.zx,
                self.xx * other.xy + self.xy * other.yy + self.xz * other.zy,
                self.xx * other.xz + self.xy * other.yz + self.xz * other.zz,
                self.yx * other.xx + self.yy * other.yx + self.yz * other.zx,
                self.yx * other.xy + self.yy * other.yy + self.yz * other.zy,
                self.yx * other.xz + self.yy * other.yz + self.yz * other.zz,
                self.zx * other.xx + self.zy * other.yx + self.zz * other.zx,
                self.zx * other.xy + self.zy * other.yy + self.zz * other.zy,
                self.zx * other.xz + self.zy * other.yz + self.zz * other.zz,
            )
        raise RuntimeError(f'No multiplication with {type(other)}')

    def __sub__(self, other: RotationType) -> Matrix:
        return Matrix(self.xx - other.xx, self.xy - other.xy, self.xz - other.xz,
                      self.yx - other.yx, self.yy - other.yy, self.yz - other.yz,
                      self.zx - other.zx, self.zy - other.zy, self.zz - other.zz)

    @staticmethod
    def from_angles(angles: Angles, degrees=True) -> RotationType:
        """Construct a rotation matrix from three angles"""
        from math import cos, sin
        from mccode_antlr.common import Expr, unary_expr
        s = [unary_expr(sin_degree if degrees else sin, 'sin', j) for j in angles]
        c = [unary_expr(cos_degree if degrees else cos, 'cos', j) for j in angles]
        z, o = Expr.float(0), Expr.float(1)
        rx = Rotation() if angles.x.is_zero else Rotation(o, z, z, z, c[0], s[0], z, -s[0], c[0])
        ry = Rotation() if angles.y.is_zero else Rotation(c[1], z, -s[1], z, o, z, s[1], z, c[1])
        rz = Rotation() if angles.z.is_zero else Rotation(c[2], s[2], z, -s[2], c[2], z, z, z, o)
        # Return the McCode convention of applying the rotations in the order x, y, z
        return rz * (ry * rx)

    def __contains__(self, value):
        return any(value in x for x in (self.xx, self.xy, self.xz, self.yx, self.yy, self.yz, self.zx, self.zy, self.zz))


class Seitz(NamedTuple):
    xx: Expr = Expr.float(1)
    xy: Expr = Expr.float(0)
    xz: Expr = Expr.float(0)
    xt: Expr = Expr.float(0)
    yx: Expr = Expr.float(0)
    yy: Expr = Expr.float(1)
    yz: Expr = Expr.float(0)
    yt: Expr = Expr.float(0)
    zx: Expr = Expr.float(0)
    zy: Expr = Expr.float(0)
    zz: Expr = Expr.float(1)
    zt: Expr = Expr.float(0)

    @classmethod
    def from_rotation(cls, rotation: Rotation):
        return cls(rotation.xx, rotation.xy, rotation.xz, Expr.float(0),
                   rotation.yx, rotation.yy, rotation.yz, Expr.float(0),
                   rotation.zx, rotation.zy, rotation.zz, Expr.float(0))

    @classmethod
    def from_vector(cls, vector: Vector):
        return cls(Expr.float(1), Expr.float(0), Expr.float(0), vector.x,
                   Expr.float(0), Expr.float(1), Expr.float(0), vector.y,
                   Expr.float(0), Expr.float(0), Expr.float(1), vector.z)

    def __str__(self):
        return f'[{str(self.rotation())}|{str(self.vector())}]'

    def rotation(self) -> Rotation:
        return Rotation(self.xx, self.xy, self.xz, self.yx, self.yy, self.yz, self.zx, self.zy, self.zz)

    def vector(self) -> Vector:
        return Vector(self.xt, self.yt, self.zt)

    def inverse(self) -> SeitzType:
        """
        Special 4x3 (flat, row-ordered) matrix inverse.

        For the special case of a _projective transformation matrix_ and acts on augmented vectors,
        where the omitted last row would be [0, 0, 0, 1]. In this special case, we can express the matrix as
            |R T|
            |0 1|
        where R is a 3x3 (rotation) matrix and T a 3-vector translation. Then we need only find the inverse of R, R^-1,
        and note that T^-1 = - R^-1 T to find the inverse of the affine projective transformation matrix.
        """
        inv_r = self.rotation().inverse()
        neg_inv_t = inv_r * self.vector()
        return Seitz(inv_r.xx, inv_r.xy, inv_r.xz, -neg_inv_t.x,
                     inv_r.yx, inv_r.yy, inv_r.yz, -neg_inv_t.y,
                     inv_r.zx, inv_r.zy, inv_r.zz, -neg_inv_t.z)

    def trace(self) -> Expr:
        """Abuse this to be the trace of the rotation part"""
        return self.xx + self.yy + self.zz

    def __mul__(self, other: SeitzType) -> SeitzType:
        """
        Multiplication in the special case of two _projective transformation matrices_ with omitted last row [0, 0, 0, 1].
        In this special case, we can express the matrix multiplication as
            |A t| |B s|  =  | AB  As+t |
            |0 1| |0 1|     |  0    1  |
        where A and B are 3x3 (rotation) matrices and t and s 3-vector translations.
        """
        so = self.rotation() * other.rotation()
        st = self.rotation() * other.vector()
        s = self.vector()
        return Seitz(so.xx, so.xy, so.xz, st.x + s.x,
                     so.yx, so.yy, so.yz, st.y + s.y,
                     so.zx, so.zy, so.zz, st.z + s.z)

    def __contains__(self, value):
        return value in self.rotation() or value in self.vector()

def cos_degree(theta_degree):
    from math import pi, cos
    return cos(theta_degree / 180 * pi)


def sin_degree(theta_degree):
    from math import pi, sin
    return sin(theta_degree / 180 * pi)


def atan2_degree(y, x):
    from math import pi, atan2
    return atan2(y, x) / pi * 180


def acos_degree(v):
    from math import pi, acos
    return acos(v) / pi * 180


def acos_value(v: Expr, degrees=True):
    from math import acos
    pi = Expr.float(180) if degrees else Expr.id('PI')
    if v.is_value(1) or v > Expr.float(1):
        return Expr.float(0)
    if v.is_value(0):
        return pi / Expr.float(2)
    if v.is_value(-1) or v < Expr.float(-1):
        return pi
    return unary_expr(acos_degree if degrees else acos, 'acos', v)


def cos_value(v: Expr, degrees=True):
    """The cosine of an angle expressed in degrees or radian"""
    from math import cos, pi
    if v.is_value(0):
        return Expr.float(1)
    pi = Expr.float(pi) if v.is_constant else Expr.id('PI')
    if (degrees and v.is_value(Expr.float(90))) or (not degrees and v.is_value(pi/Expr.float(2))):
        return Expr.float(0)
    if (degrees and v.is_value(Expr.float(180))) or (not degrees and v.is_value(pi)):
        return Expr.float(-1)
    return unary_expr(cos_degree if degrees else cos, 'cos', v)


def sin_value(v: Expr, degrees=True):
    from math import sin, pi
    if v.is_value(0):
        return Expr.float(0)
    pi = Expr.float(pi) if v.is_constant else Expr.id('PI')
    if (degrees and v.is_value(Expr.float(90))) or (not degrees and v.is_value(pi / Expr.float(2))):
        return Expr.float(1)
    if (degrees and v.is_value(Expr.float(180))) or (not degrees and v.is_value(-pi / Expr.float(2))):
        return Expr.float(-1)
    return unary_expr(sin_degree if degrees else sin, 'sin', v)


def atan2_value(va: Expr, vb: Expr, degrees=True):
    from math import atan2
    zero = Expr.float(0)
    full = Expr.float(180) if degrees else Expr.id('PI')
    half = full / Expr.float(2)
    if va.is_value(0):
        if vb.is_op:
            return [zero, full]
        return [zero if vb > zero else full]
    if vb.is_value(0):
        if va.is_op:
            return [-half, half]
        return [half if va > zero else -half]
    return [binary_expr(atan2_degree if degrees else atan2, 'atan2', va, vb)]


def cos_from_sin_value(v: Expr):
    from ..common import UnaryOp
    if v.is_value(0):
        return Expr.float(1)
    if v.is_value(1):
        return Expr.float(0)
    if v.is_op and isinstance(v.expr[0], UnaryOp):
        if '-' == v.expr[0].op:
            return cos_from_sin_value(Expr(v.expr[0].value))
        if 'sin' == v.expr[0].op:
            return Expr(UnaryOp('cos', v.expr[0].value))
    return sqrt_value(Expr.float(1) - v * v)


def sqrt_value(v: Expr):
    from math import sqrt
    return unary_expr(sqrt, 'sqrt', v)


def degree_to_radian(v: Expr):
    from math import pi
    if v.is_id:
        if v == Expr.id('PI'):
            raise RuntimeError(f'Convert {v} to radian')
        return v * (Expr.id('PI') / Expr.float(180))
    return v * Expr.float(pi / 180)


def _rotation_angles_to_axes_coordinates(rotated: Angles, degrees=True):
    cx, cy, cz = [cos_value(r if degrees else degree_to_radian(r), degrees=degrees) for r in rotated]
    sx, sy, sz = [sin_value(r if degrees else degree_to_radian(r), degrees=degrees) for r in rotated]
    # Rotation matrices following the McCode first x then y then z method of applying rotations.
    # The 3x3 rotation matrix part (which rotates the *axes* of a coordinate system):
    axes = (cy * cz, sx * sy * cz + cx * sz, sx * sz - cx * sy * cz,
            -cy * sz, cx * cz - sx * sy * sz, sx * cz + cx * sy * sz,
            sy, -sx * cy, cx * cy)
    # The coordinates of the same system rotate the opposite way, but still in the same order
    # (All sin terms gain a negative sign)
    coordinates = (cy * cz, sx * sy * cz - cx * sz, sx * sz + cx * sy * cz,
                   cy * sz, cx * cz + sx * sy * sz, -sx * cz + cx * sy * sz,
                   -sy, sx * cy, cx * cy)
    return axes, coordinates


def axes_euler_angles(m: Rotation, degrees) -> Angles:
    """Return the angles vector which could have been used to produce this 'axes' rotation matrix"""
    rxs = atan2_value(-m.zy, m.zz, degrees=degrees)
    rys = atan2_value(m.zx, cos_from_sin_value(m.zx), degrees=degrees)
    rzs = atan2_value(-m.yx, m.xx, degrees=degrees)
    rx = rxs[0] if len(rxs) == 1 else rxs[0]
    ry = rys[0] if len(rys) == 1 else rys[0]
    rz = rzs[0] if len(rzs) == 1 else rzs[0]
    return Angles(rx, ry, rz)


OrientationPartType = TypeVar('OrientationPartType', bound='OrientationPart')


@dataclass
class Part:
    """The Seitz matrix part of any arbitrary projective affine transformation"""
    _axes: Seitz = field(default_factory=Seitz)

    def __post_init__(self):
        """If this is not defined, the subclass' __post_init__ may not be called"""
        # (On one system, the absence of this method did not affect the calling of the subclass post_init)
        pass

    @property
    def is_translation(self):
        return any(not p.is_zero for p in (self._axes[3], self._axes[7], self._axes[11]))

    @property
    def is_rotation(self):
        # The first condition _should_ always be true -- the second is only true if this is not the identity matrix
        if round((self._axes.inverse() * self._axes).trace(), 12) == Expr.float(3.):
            return round(self._axes.trace(), 12) != Expr.float(3.)
        log.info(f'Not a rotation matrix: {self._axes}')
        return False

    @property
    def is_identity(self):
        return not self.is_translation and self._axes.trace() == Expr.float(3.)

    @property
    def rotation_axis_angle(self) -> tuple[Vector, Expr, str]:
        from numpy.linalg import eig
        from numpy import array, argmin, sqrt, real, imag, conj, sum
        if not self.is_constant:
            raise RuntimeError('Not possible to determine the rotation axis and angle of a variable general transform')
        # For the matrix A in cross(axis, v) = A v,
        # If flat is _axes, then R = 1 - A sin(angle) + A^2 (1-cos(angle))
        #  if _coordinates, then R = 1 + A sin(angle) + A^2 (1-cos(angle))
        # where the difference in sign is due to _coordinates rotating points and _axes rotating the axes instead.
        flat = self.rotation('coordinates')
        matrix = [[flat.xx.value, flat.xy.value, flat.xz.value],
                  [flat.yx.value, flat.yy.value, flat.yz.value],
                  [flat.zx.value, flat.zy.value, flat.zz.value]]
        dd, vv = eig(array(matrix))
        # The eigenvalues, dd, are (1+0j, a+bj, a-bj) of which we want 1+0j.
        axis = vv[:, argmin(sqrt(real(conj(dd-1) * (dd-1))))]
        if sum(imag(axis)) != 0:
            log.warn(f'Imaginary rotation axis {real(axis)} + j {imag(axis)}')
        axis = real(axis)
        cos_angle = (matrix[0][0] + matrix[1][1] + matrix[2][2] - 1) / 2
        if abs(cos_angle) > 1:
            log.warn(f'Invalid cos(angle) {cos_angle} for {self}')
            cos_angle = 1 if cos_angle > 0 else -1
        angle = Expr.float(acos_degree(cos_angle))
        axis = Vector(Expr.float(axis[0]), Expr.float(axis[1]), Expr.float(axis[2]))

        # Check that the rotation matrix signs all match:
        a = axis.cross_matrix()
        r = Matrix.eye() + a * sin_value(angle, True) + (a * a) * (Expr.float(1) - cos_value(angle, True))
        # and flip the sign of angle if not:
        return axis, -angle if any(flat[index] * r[index] < Expr.float(0) for index in range(9)) else angle, 'degrees'

    def __str__(self):
        x = [f'{x}' for x in self._axes]
        widths = [max(len(x[i+4*j]) for j in range(3)) for i in range(4)]
        line_fmt = [f'>{w+2:d}s' for w in widths]
        lines = '\n'.join(' '.join(f'{x[i+4*j]:{f:s}}' for i, f in zip(range(4), line_fmt)) for j in range(3))
        return lines

    def __repr__(self):
        inner = ','.join(f'{x}' for x in self._axes)
        return f'{self.__class__.__name__}({inner})'

    @property
    def axes(self) -> Seitz:
        return self._axes

    @property
    def coordinates(self) -> Seitz:
        return self._axes.inverse()

    def seitz(self, which=None) -> Seitz:
        return self.axes if which is None or 'axes' in which else self.coordinates

    def position(self, which=None) -> Vector:
        return self.seitz(which).vector()

    def rotation(self, which=None) -> Rotation:
        return self.seitz(which).rotation()

    def angles(self, which=None, degrees=True) -> Angles:
        return axes_euler_angles(self.seitz('axes').rotation(), degrees)

    @property
    def all_values(self) -> bool:
        return all(x.has_value for x in self._axes)

    @property
    def is_constant(self) -> bool:
        return all(x.is_constant for x in self._axes)

    def inverse(self: OrientationPartType) -> OrientationPartType:
        return Part(self._axes.inverse())

    def __mul__(self: OrientationPartType, other: OrientationPartType) -> OrientationPartType:
        if not isinstance(other, Part):
            raise RuntimeError('Multiplication only defined for two orientation parts')
        return Part(self.axes * other.axes)

    def specialization(self):
        if not self.is_rotation:
            return TranslationPart(v=self.position())
        elif not self.is_translation:
            axis, angles, units = self.rotation_axis_angle
            if axis == Vector(Expr.float(1), Expr.float(0), Expr.float(0)):
                return RotationX(v=angles, degrees=units == 'degrees')
            elif axis == Vector(Expr.float(0), Expr.float(1), Expr.float(0)):
                return RotationY(v=angles, degrees=units == 'degrees')
            elif axis == Vector(Expr.float(0), Expr.float(0), Expr.float(1)):
                return RotationZ(v=angles, degrees=units == 'degrees')
        return self

    def __eq__(self, other: OrientationPartType):
        # Two orientation part objects are equal if and only if their axes are equal
        return self.axes == other.axes

    def __contains__(self, value):
        return value in self._axes


@dataclass
class TranslationPart(Part):
    """A specialization to the translation-only part of a projective affine transformation"""
    v: Vector = field(default_factory=Vector)

    def __str__(self):
        return f'({self.v[0]}, {self.v[1]}, {self.v[2]}) [0, 0, 0]'

    def __repr__(self):
        return f'{self.__class__.__name__}({self.v})'

    @property
    def is_constant(self):
        return all(x.has_value for x in self.v)

    @property
    def is_translation(self):
        return True

    @property
    def is_rotation(self):
        return False

    @property
    def rotation_axis_angle(self) -> tuple[Vector, Expr, str]:
        raise RuntimeError('Not possible to determine the rotation axis and angle of a translation')

    def __post_init__(self):
        z, o = Expr.float(0), Expr.float(1)
        self._axes = Seitz(o, z, z, self.v[0], z, o, z, self.v[1], z, z, o, self.v[2])

    def inverse(self):
        return TranslationPart(v=Vector(-self.v[0], -self.v[1], -self.v[2]))

    def position(self, which=None) -> Vector:
        return self.v

    def angles(self, which=None, degrees=True) -> Angles:
        z = Expr.float(0)
        return Angles(z, z, z)

    def __contains__(self, value):
        return value in self.v


@dataclass
class RotationPart(Part):
    """A specialization to the rotation-only part of a projective affine transformation"""
    v: Expr = Expr.float(0)
    degrees: bool = True

    def __post_init__(self):
        """If this is not defined, the subclass' __post_init__ will not be called"""
        # On one system the absence of this method prevented the subclass' __post_init__ from being called
        pass

    @property
    def is_constant(self):
        return self.v.has_value

    @property
    def is_translation(self):
        return False

    @property
    def is_rotation(self):
        return not self.v.is_zero

    @property
    def rotation_axis(self) -> Vector:
        return Vector(Expr.float(0), Expr.float(0),  Expr.float(0))

    @property
    def rotation_axis_angle(self) -> tuple[Vector, Expr, str]:
        return self.rotation_axis, self.v, 'degrees' if self.degrees else 'radian'

    def _cos_sin_one_zero(self):
        r = self.v if self.degrees else degree_to_radian(self.v)
        return cos_value(r), sin_value(r), Expr.float(1), Expr.float(0)

    def position(self, which=None) -> Vector:
        z = Expr.float(0)
        return Vector(z, z, z)

    def __str__(self):
        return f'(0, 0, 0) [? ? ?; {self.v}]'

    def __repr__(self):
        return f'{self.__class__.__name__}({self.v})'

    def __contains__(self, value):
        return value in self.v


class RotationX(RotationPart):
    """A specialization to the rotation-around-X part of a projective affine transformation"""
    def __post_init__(self):
        c, s, o, z = self._cos_sin_one_zero()
        self._axes = Seitz(o, z, z, z, z, c, s, z, z, -s, c, z)

    def inverse(self):
        return RotationX(v=-self.v, degrees=self.degrees)

    def angles(self, which=None, degrees=True) -> Angles:
        z = Expr.float(0)
        return Angles(self.v, z, z)

    def __str__(self):
        return f'(0, 0, 0) [{self.v}, 0, 0]'

    @property
    def rotation_axis(self) -> Vector:
        return Vector(Expr.float(1), Expr.float(0), Expr.float(0))


class RotationY(RotationPart):
    """A specialization to the rotation-around-Y part of a projective affine transformation"""
    def __post_init__(self):
        c, s, o, z = self._cos_sin_one_zero()
        self._axes = Seitz(c, z, -s, z, z, o, z, z, s, z, c, z)

    def inverse(self):
        return RotationY(v=-self.v, degrees=self.degrees)

    def angles(self, which=None, degrees=True) -> Angles:
        z = Expr.float(0)
        return Angles(z, self.v, z)

    def __str__(self):
        return f'(0, 0, 0) [0, {self.v}, 0]'

    @property
    def rotation_axis(self) -> Vector:
        return Vector(Expr.float(0), Expr.float(1), Expr.float(0))


class RotationZ(RotationPart):
    """A specialization to the rotation-around-Z part of a projective affine transformation"""
    def __post_init__(self):
        c, s, o, z = self._cos_sin_one_zero()
        self._axes = Seitz(c, s, z, z, -s, c, z, z, z, z, o, z)

    def inverse(self):
        return RotationZ(v=-self.v, degrees=self.degrees)

    def angles(self, which=None, degrees=True) -> Angles:
        z = Expr.float(0)
        return Angles(z, z, self.v)

    def __str__(self):
        return f'(0, 0, 0) [0, 0, {self.v}]'

    @property
    def rotation_axis(self) -> Vector:
        return Vector(Expr.float(0), Expr.float(0), Expr.float(1))


PartsType = TypeVar('PartsType', bound='Parts')


@dataclass
class Parts:
    """A list of unresolved or partially resolved successive projective affine transformation(s)"""
    _stack: tuple[Part, ...] = field(default_factory=tuple)

    def __str__(self):
        inner = ','.join(str(x) for x in self._stack) if len(self._stack) else ''
        return f'Parts<{inner}>'

    def __repr__(self):
        inner = ','.join(repr(x) for x in self._stack) if len(self._stack) else ''
        return f'Parts<{inner}>'

    def stack(self):
        return self._stack

    def _copy(self, deep: bool = True) -> tuple[Part]:
        if deep:
            from copy import deepcopy
            return tuple([deepcopy(x) for x in self._stack])
        return tuple(x for x in self._stack)

    def copy(self, deep: bool = True) -> PartsType:
        return Parts(self._copy(deep))

    def __add__(self, other) -> PartsType:
        out = self._copy()
        if isinstance(other, Parts):
            out += other.stack()
        elif isinstance(other, Part):
            out += (other,)
        else:
            raise ValueError(f'__add__ undefined for {type(self)} and {type(other)}')
        return Parts(out)

    def __sub__(self, other) -> PartsType:
        out = self._copy()
        if isinstance(other, Parts):
            out += tuple(reversed([x.inverse() for x in other.stack()]))
        elif isinstance(other, Part):
            out += (other.inverse(),)
        else:
            raise ValueError(f'__sub__ undefined for {type(self)} and {type(other)}')
        return Parts(out)

    @classmethod
    def from_at_rotated(cls, at: Vector, rotated: Angles, degrees=True):
        """Emulate the McCode ordered-4-part-transformation:
            1. Translate to the specified position
            2. Rotate around the current X axis
            3. Rotate around the resulting Y axis
            4. Rotate around the resulting Z axis
        Such a method can make some orientations difficult to express, and the McCode solution
        is to chain successive ordered transformations with some parts set to zero.
        """
        # The order of individual OrientationPart objects in this tuple is paramount
        s = tuple()
        if not rotated[0].is_zero:
            s += (RotationX(v=rotated[0], degrees=degrees), )
        if not rotated[1].is_zero:
            s += (RotationY(v=rotated[1], degrees=degrees), )
        if not rotated[2].is_zero:
            s += (RotationZ(v=rotated[2], degrees=degrees), )
        if not all(x.is_zero for x in at):
            s += (TranslationPart(v=at),)
        return cls(s)

    @classmethod
    def from_dependent_chain(cls, dep: PartsType, at: Vector = None, rotated: Angles = None, degrees=True, copy=True):
        """
        If it is beneficial to rotate around Z before rotating around the resulting X,
        one can have successive components which are rotated (0,0,rz) (rx,0,0).
        By using the from_dependent_chain method these two projective transformations can be combined
        in the correct order

        :param dep: The projective transformation list which proceeds the newly added one(s)
        :param at: The translational part of the new projective transformation
        :param rotated: The angles of the three successive rotational parts
        :param degrees: Whether the angles are in degrees or radian
        :param copy: Whether the proceeding list should be copied
        :return: A OrientationParts object representing the chained operations
        """
        stack = () if dep is None else dep._copy(deep=copy)
        stack += cls.from_at_rotated(at or Vector(), rotated or Angles(), degrees).stack()
        return cls(stack).reduce()

    def inverse(self):
        """Return the inverse transformation of the full chain
        For operations Z Y X T R, the inverse is (Z Y X T R)^-1 which has the property
            (Z Y X T R)^-1 Z Y X T R = 1
        from inspection this can only be true if the inverse applies the inverse of each
        projective-transformation-part in reverse order
            R^-1 T^-1 X^-1 Y^-1 Z^-1 Z Y X T R = 1
        """
        return Parts(tuple(x.inverse() for x in reversed(self._stack)))

    def reduce(self):
        """Combine all successive constant projective-transformation parts in the chain
        If there are parts X(x), Y(y), Z(z) that functions of an unknown (runtime defined)
        parameter and any number of constant-valued parts Ci interspersed like
            C1 C2 C3 X(x) C3 C4 Y(y) C5 Z(z) C6 ... CN
        return the reduced list
            C123 X(x) C34 Y(y) C5 Z(z) C6N
        where, e.g., C123 = C1 C2 C3
        """
        if not len(self._stack):
            return Parts()

        reduced = [self._stack[0]]
        for top in self._stack[1:]:
            if reduced[-1].is_constant and reduced[-1].is_identity:
                # Any combination which results in the identity matrix can be removed
                reduced[-1] = top
            elif reduced[-1].is_constant and top.is_constant:
                reduced[-1] = top * reduced[-1]
            else:
                reduced.append(top)
                
        if len(reduced) > 1 and reduced[-1].is_constant and reduced[-1].is_identity:
            reduced.pop()

        for x in reduced:
            if not isinstance(x, Part):
                raise RuntimeError('Non-OrientationPart in reduced list')
        # See if any entries can be replaced by specializations
        # return OrientationParts(tuple(x.specialization() for x in reduced))
        return Parts(tuple(x for x in reduced if isinstance(x, Part)))

    def resolve(self):
        """Fully combine all parts of the chain, may raise an error if such a proceedure is disallowed"""
        if not len(self._stack):
            return Part()

        resolved = self._stack[0]
        for work in self._stack[1:]:
            resolved = work * resolved
        return resolved

    def position(self, which=None):
        return self.resolve().position(which=which)

    def rotation(self, which=None):
        return self.resolve().rotation(which=which)

    def __contains__(self, value_name):
        return any(value_name in element for element in self._stack)


OrientType = TypeVar('OrientType', bound='Orient')


@dataclass
class Orient:
    """Un-evaluated lists of dependent operations that position and orient an object in a coordinate system"""
    _position: Parts = field(default_factory=Parts)
    _rotation: Parts = field(default_factory=Parts)
    _degrees: bool = True

    def __str__(self):
        return f'Orient<{self._position}, {self._rotation}>'

    def __repr__(self):
        return f'Orient<{repr(self._position)}, {repr(self._rotation)}>'

    @property
    def degrees(self):
        return self._degrees

    def get_raw_position(self):
        return self._position

    def get_raw_rotation(self):
        return self._rotation

    @classmethod
    def from_dependent_orientations(cls, rel: OrientType, at: Vector, rot: OrientType, angles: Angles,
                                    degrees=True, copy=True):
        # if 'at' was defined RELATIVE to another component, the vector is *in that component's coordinate system*
        at = Vector(*at) if isinstance(at, tuple) else at
        angles = Angles(*angles) if isinstance(angles, tuple) else angles
        rel = rel or Orient()
        pos = Parts((TranslationPart(v=rel.position() + rel.rotation('coordinates') * at),))
        rot = Parts.from_dependent_chain(rot and rot.rotation_parts(), rotated=angles, degrees=degrees, copy=copy)
        return cls(pos, rot, degrees)

    @classmethod
    def from_dependent_orientation(cls, dep: OrientType, at: Vector, angles: Angles, degrees=True, copy=True):
        dep = dep or Orient()
        at = Vector(*at) if isinstance(at, tuple) else at
        angles = Angles(*angles) if isinstance(angles, tuple) else angles
        pos = Parts((TranslationPart(v=dep.position() + dep.rotation('coordinates') * at),))
        rot = Parts.from_dependent_chain(dep.rotation_parts(), rotated=angles, degrees=degrees, copy=copy)
        return cls(pos, rot, degrees)

    def angles(self, which=None) -> Angles:
        return self._rotation.resolve().angles(which=which)

    def position(self, which=None) -> Vector:
        return self._position.position(which=which)

    def rotation(self, which=None) -> Rotation:
        return self._rotation.rotation(which=which)

    def position_parts(self) -> Parts:
        return self._position.reduce()

    def rotation_parts(self, which=None) -> Parts:
        return self._rotation.reduce()

    def inverse(self):
        return Orient(self._position.inverse(), self._rotation.inverse())

    def reduce(self):
        return Orient(self._position.reduce(), self._rotation.reduce())

    def combine(self) -> Parts:
        # The position represents a coordinate and the rotation represents the axes of that coordinate system
        # So even though we apply the translation first, the rotation operations need to appear first in the
        # combined set of operations:
        comb = self._rotation.copy().stack()
        comb += self._position.copy().stack()
        return Parts(comb)

    def __add__(self, other):
        if isinstance(other, Orient):
            return Orient(self._position + other._position, self._rotation + other._rotation)
        elif isinstance(other, (Part, Parts)):
            return Orient(self._position + other, self._rotation + other)
        else:
            raise ValueError(f"__add__ undefined for DependentOrientation and {type(other)}")

    def __sub__(self, other):
        if isinstance(other, Orient):
            pos = self._position - other._position
            return Orient(other._rotation + pos, self._rotation - other._rotation)
        elif isinstance(other, (Part, Parts)):
            return Orient(self._position - other, self._rotation - other)
        else:
            raise ValueError(f"__sub__ undefined for DependentOrientation and {type(other)}")

    def __contains__(self, value_name):
        return value_name in self._position or value_name in self._rotation



from pathlib import Path
from typing import Union
from mccode_antlr.instr import Instr
from mccode_antlr.reader import Registry


def parse_mccode_instr_parameters(contents: str):
    from antlr4 import CommonTokenStream, InputStream
    from mccode_antlr.grammar import McInstrParser, McInstrLexer
    from mccode_antlr.instr import InstrParametersVisitor
    parser = McInstrParser(CommonTokenStream(McInstrLexer(InputStream(contents))))
    visitor = InstrParametersVisitor()
    return visitor.visitProg(parser.prog())


def parse_mccode_instr(contents: str, registry: Registry, source: str = None) -> Instr:
    from antlr4 import CommonTokenStream, InputStream
    from mccode_antlr.grammar import McInstrParser, McInstrLexer
    from mccode_antlr.instr import InstrVisitor
    from mccode_antlr.reader import Reader
    parser = McInstrParser(CommonTokenStream(McInstrLexer(InputStream(contents))))
    registries = [registry]
    reader = Reader(registries=registries)
    visitor = InstrVisitor(reader, source or '<string>')
    instr = visitor.visitProg(parser.prog())
    instr.flags = tuple(reader.c_flags)
    instr.registries = tuple(registries)
    return instr


def parse_mcstas_instr(contents: str) -> Instr:
    from mccode_antlr.reader import MCSTAS_REGISTRY
    return parse_mccode_instr(contents, MCSTAS_REGISTRY)


def parse_mcxtrace_instr(contents: str) -> Instr:
    from mccode_antlr.reader import MCXTRACE_REGISTRY
    return parse_mccode_instr(contents, MCXTRACE_REGISTRY)


def load_mccode_instr(filename: Union[str, Path], registry: Registry) -> Instr:
    return parse_mccode_instr(Path(filename).read_text(), registry, source=str(filename))


def load_mcstas_instr(filename: Union[str, Path]) -> Instr:
    from mccode_antlr.reader import MCSTAS_REGISTRY
    return load_mccode_instr(filename, MCSTAS_REGISTRY)


def load_mcxtrace_instr(filename: Union[str, Path]) -> Instr:
    from mccode_antlr.reader import MCXTRACE_REGISTRY
    return load_mccode_instr(filename, MCXTRACE_REGISTRY)



from zenlog import log

def rebuild_language(grammar_file, verbose=False):
    from pathlib import Path
    from subprocess import Popen, PIPE
    from antlr4_tool_runner import initialize_paths, get_version_arg, install_jre_and_antlr

    if not isinstance(grammar_file, Path):
        grammar_file = Path(grammar_file)

    # Do the McCode*.py files not exist? Or are they older than McCode.g4?
    # Then build them using ANTLR4:
    args = ['-Dlanguage=Python3', str(grammar_file), '-listener', '-visitor', '-o', str(grammar_file.parent)]

    # The following copies the implementation of antlr4_tool_runner.tool, which pulls `args` from the system argv list
    # Setup:
    initialize_paths()
    args, version = get_version_arg(args)
    jar, java = install_jre_and_antlr(version)
    # Call antlr4
    p = Popen([java, '-cp', jar, 'org.antlr.v4.Tool'] + args, stdout=PIPE, stderr=PIPE)
    out, err = [x.decode('UTF-8') for x in p.communicate()]
    if err:
        print(err, end='')
    if out:
        print(out, end='')


def language_present_and_up_to_date(grammar_file, newest, verbose=False):
    from pathlib import Path
    if not isinstance(grammar_file, Path):
        grammar_file = Path(grammar_file)

    if not grammar_file.exists():
        raise RuntimeError(f"The grammar file {grammar_file} does not exist.")

    root = grammar_file.parent
    lexer_file = root.joinpath(f'{grammar_file.stem}Lexer.py')
    parser_file = root.joinpath(f'{grammar_file.stem}Parser.py')
    listener_file = root.joinpath(f'{grammar_file.stem}Listener.py')
    visitor_file = root.joinpath(f'{grammar_file.stem}Visitor.py')

    generated_files = lexer_file, parser_file, listener_file, visitor_file

    if not all(x.exists() for x in generated_files):
        if verbose:
            log.info(f'Not all language files exist for {grammar_file}')
        return False

    if any(x.stat().st_mtime < newest for x in generated_files):
        if verbose:
            log.info(f'Not all language files up-to-date for {grammar_file}')
        return False

    if verbose:
        log.info(f'Language files for {grammar_file} are up-to-date')
    return True


def _ensure_antlr_files_up_to_date_on_import(grammar, deps=None, verbose=False):
    """Run on import of this (sub)module. Ensure the ANTLR parsed language files are up-to-date."""
    from pathlib import Path
    grammar_file = Path(__file__).parent.joinpath(f'{grammar}.g4')

    newest = grammar_file.stat().st_mtime
    if deps:
        for dep in deps:
            newest = max(newest, Path(__file__).parent.joinpath(f'{dep}.g4').stat().st_mtime)

    if not language_present_and_up_to_date(grammar_file, newest, verbose=verbose):
        rebuild_language(grammar_file, verbose=verbose)



from .reader import Reader
from .registry import (Registry, LocalRegistry, RemoteRegistry, ModuleRemoteRegistry, GitHubRegistry,
                       MCSTAS_REGISTRY, MCXTRACE_REGISTRY, LIBC_REGISTRY, FIXED_LIBC_REGISTRY)

__all__ = [
    'Reader',
    'Registry',
    'LocalRegistry',
    'RemoteRegistry',
    'ModuleRemoteRegistry',
    'GitHubRegistry',
    'MCSTAS_REGISTRY',
    'MCXTRACE_REGISTRY',
    'LIBC_REGISTRY',
    'FIXED_LIBC_REGISTRY',
]



import ast
from dataclasses import dataclass
from typing import Union
from enum import Enum
from zenlog import log

# class NameReplacer(ast.NodeTransformer):
#     def __init__(self, **values):
#         self.values = values
#
#     def visit_Name(self, node):
#         if node.id in self.values:
#             return ast.Constant(self.values[node.id])
#         return node
#
#
# @dataclass
# class Expr:
#     expr: str = ""
#
#     @property
#     def ids(self):
#         from ast import walk, parse, Name
#         tree = parse(self.expr)
#         return list(set(x.id for x in walk(tree) if isinstance(x, Name)))
#
#     def eval(self, **values):
#         from ast import parse, fix_missing_locations, dump, literal_eval
#         if any(name not in values for name in self.ids):
#             raise RuntimeError(f"Not all of {self.ids} provided as keyword=value arguments")
#         tree = NameReplacer(**values).visit(parse(self.expr))
#         fixed = fix_missing_locations(tree)
#         print(eval(fixed))
#         print(dump(fixed))
#         obj = compile(fixed, '', 'exec')
#         eval(obj)


class ObjectType(Enum):
    value = 1
    initializer_list = 2
    identifier = 3
    function = 4
    parameter = 5

    def __str__(self):
        return self.name

    @staticmethod
    def from_name(name):
        if 'value' in name:
            return ObjectType.value
        if 'initializer_list' in name:
            return ObjectType.initializer_list
        if 'identifier' in name:
            return ObjectType.identifier
        if 'function' in name:
            return ObjectType.function
        if 'parameter' in name:
            return ObjectType.parameter
        raise RuntimeError(f"No known conversion from non-enumerated object type {name}")

    @property
    def is_id(self):
        return self == ObjectType.identifier

    @property
    def is_parameter(self):
        return self == ObjectType.parameter

    @property
    def is_function(self):
        return self == ObjectType.function


class ShapeType(Enum):
    unknown = 0
    scalar = 1
    vector = 2

    @property
    def mccode_c_type(self):
        return '*' if self.is_vector else ''

    def compatible(self, other):
        return self == ShapeType.unknown or other == ShapeType.unknown or self == other

    @property
    def is_scalar(self):
        return self == ShapeType.scalar

    @property
    def is_vector(self):
        return self == ShapeType.vector

    def __str__(self):
        return self.name

    @staticmethod
    def from_name(name):
        if 'vector' in name:
            return ShapeType.vector
        if 'scalar' in name:
            return ShapeType.scalar
        return ShapeType.unknown


class DataType(Enum):
    undefined = 0
    float = 1
    int = 2
    str = 3

    def compatible(self, other):
        if self == DataType.undefined or other == DataType.undefined or self == other:
            return True
        if (self == DataType.float and other == DataType.int) or (self == DataType.int and other == DataType.float):
            return True
        return False

    # promotion rules:
    def __add__(self, other):
        if self == DataType.undefined:
            return other
        if other == DataType.undefined:
            return self
        if self == other:
            return self
        if (self == DataType.float and other == DataType.int) or (self == DataType.int and other == DataType.float):
            return DataType.float
        return DataType.str

    __sub__ = __add__
    __mul__ = __add__

    def __truediv__(self, other):
        if self == DataType.str or other == DataType.str:
            raise RuntimeError('Division of strings is undefined')
        return DataType.float

    def __floordiv__(self, other):
        return DataType.int

    @classmethod
    def from_name(cls, name):
        if 'double' in name or 'float' in name:
            return cls.float
        if 'int' in name:
            return cls.int
        if 'char' in name or 'string' in name or 'str' in name:
            return cls.str
        return cls.undefined

    @property
    def name(self):
        if self == DataType.int:
            return 'int'
        if self == DataType.float:
            return 'float'
        if self == DataType.str:
            return 'str'
        return 'undefined'

    @property
    def is_int(self):
        return self == DataType.int

    @property
    def is_float(self):
        return self == DataType.float

    @property
    def is_str(self):
        return self == DataType.str

    @property
    def mccode_c_type(self):
        if self == DataType.float:
            return "double"
        if self == DataType.int:
            return "int"
        if self == DataType.str:
            return "char *"
        raise RuntimeError(f"No known conversion from non-enumerated data type {self}")


def _comb(f, s: list):
    return ','.join(f(x) for x in s)


def _fmt_comb(fmt, s: list):
    return ','.join(f'{x:{fmt}}' for x in s)


class Op:
    def __init__(self):
        self.data_type = DataType.undefined
        self.style = 'C'  # there should be a better way to do this

    def __hash__(self):
        return hash(str(self))

    def __int__(self):
        raise RuntimeError('No conversion to int for operations')

    @property
    def is_op(self):
        return True

    @property
    def is_zero(self):
        return False

    @property
    def is_id(self):
        return False

    @property
    def is_constant(self):
        return False

    @property
    def is_parameter(self):
        return False

    @property
    def is_str(self):
        return self.data_type.is_str

    def is_value(self, other):
        return self == other

    @property
    def is_scalar(self):
        return False

    @property
    def is_vector(self):
        return False

    @property
    def vector_known(self):
        return False

    def __add__(self, other):
        if other.is_zero:
            return self
        return BinaryOp('+', self, other)

    def __sub__(self, other):
        if other.is_zero:
            return self
        return BinaryOp('-', self, other)

    def __mul__(self, other):
        if other.is_zero:
            return other
        if other.is_value(1):
            return self
        if other.is_value(-1):
            return -self
        return BinaryOp('*', self, other)

    def __truediv__(self, other):
        if other.is_zero:
            raise RuntimeError('Division by zero')
        if other.is_value(1):
            return self
        if other.is_value(-1):
            return -self
        return BinaryOp('/', self, other)

    def __floordiv__(self, other):
        if other.is_zero:
            raise RuntimeError('Division by zero')
        if other.is_value(1):
            return self
        if other.is_value(-1):
            return -self
        return BinaryOp('//', self, other)

    def __neg__(self):
        return UnaryOp('-', self)

    def __pos__(self):
        return self

    def __abs__(self):
        return UnaryOp('abs', self)

    def as_type(self, pdt):
        raise NotImplementedError()


class TrinaryOp(Op):
    def __init__(self, op, first, second, third):
        super().__init__()
        self.op = op
        first, second, third = [x.expr if isinstance(x, Expr) else x for x in (first, second, third)]
        first, second, third = [[x] if not isinstance(x, list) else x for x in (first, second, third)]
        self.first = first
        self.second = second
        self.third = third
        # In C the data types for `if_true` and `if_false` in
        #   `(logical test) ? if_true : if_false;`
        # _must_ be the same. Python is more flexible, but we need to at least have a promotable common type
        # for compatibility with C.
        left_data_types = list(dict.fromkeys(x.data_type for x in second))
        right_data_types = list(dict.fromkeys(x.data_type for x in third))
        if len(left_data_types) != 1 or len(right_data_types) != 1:
            raise RuntimeError('Multiple data types in one value not supported')
        self.data_type = left_data_types[0] + right_data_types[0]

    def as_type(self, pdt):
        first = [x.as_type(pdt) for x in self.first]
        second = [x.as_type(pdt) for x in self.second]
        third = [x.as_type(pdt) for x in self.third]
        return TrinaryOp(self.op, first, second, third)

    def _str_repr_(self, first, second, third):
        if '__trinary__' == self.op:
            if self.style == 'C':
                return f'{first} ? {second} : {third}'
            return f'{second} if {first} else {third}'

    def __str__(self):
        return self._str_repr_(_comb(str, self.first), _comb(str, self.second), _comb(str, self.third))

    def __format__(self, format_spec):
        return self._str_repr_(_fmt_comb(format_spec, self.first), _fmt_comb(format_spec, self.second), _fmt_comb(format_spec, self.third))

    def __repr__(self):
        return self._str_repr_(_comb(repr, self.first), _comb(repr, self.second), _comb(repr, self.third))

    def __eq__(self, other):
        if not isinstance(other, TrinaryOp):
            return False
        return self.op == other.op and self.first == other.first and self.second == other.second and self.third == other.third

    @property
    def is_scalar(self):
        return len(self.first) == 1 and len(self.second) == 1 and len(self.third) == 1\
            and self.first[0].is_scalar and self.second[0].is_scalar and self.third[0].is_scalar

    @property
    def is_vector(self):
        return len(self.first) == 1 and len(self.second) == 1 and len(self.third) == 1 \
            and self.first[0].is_vector and self.second[0].is_vector and self.third[0].is_vector

    @property
    def vector_known(self):
        return len(self.first) == 1 and len(self.second) == 1 and len(self.third) == 1 \
            and self.first[0].vector_known and self.second[0].vector_known and self.third[0].vector_known

    def simplify(self):
        f, s, t = [[x.simplify() for x in y] for y in (self.first, self.second, self.third)]
        if self.op == '__trinary__' and len(f) == 1:
            if f[0].is_value(True) and not f[0].is_value(False):
                return Expr(s)
            if f[0].is_value(False) and not f[0].is_value(True):
                return Expr(t)
        return TrinaryOp(self.op, f, s, t)

    def evaluate(self, known: dict):
        first, second, third = [[x.evaluate(known) for x in y] for y in (self.first, self.second, self.third)]
        return TrinaryOp(self.op, first, second, third).simplify()

    def depends_on(self, name: str):
        return any(any(x.depends_on(name) for x in y) for y in (self.first, self.second, self.third))

    def copy(self):
        return TrinaryOp(self.op, self.first, self.second, self.third)

    def __contains__(self, value):
        # first, second, and third are lists, so we need to check each element to avoid using __eq__
        return any([any(value in x for x in y) for y in (self.first, self.second, self.third)])

    def verify_parameters(self, instrument_parameter_names: list[str]):
        for lr in (self.first, self.second, self.third):
            for x in lr:
                x.verify_parameters(instrument_parameter_names)


class BinaryOp(Op):
    def __init__(self, op, left, right):
        super().__init__()
        if isinstance(left, Expr):
            left = left.expr
        if isinstance(right, Expr):
            right = right.expr
        if not isinstance(left, list):
            left = [left]
        if not isinstance(right, list):
            right = [right]
        self.op = op
        self.left = left
        self.right = right
        if op in ('__call__', '__struct_access__', '__pointer_access__', '__getitem__'):
            data_type = DataType.undefined
        else:
            left_data_types = list(dict.fromkeys(x.data_type for x in left))
            right_data_types = list(dict.fromkeys(x.data_type for x in right))
            if len(left_data_types) != 1 or len(right_data_types) != 1:
                raise RuntimeError('Multiple data types in one value not supported')
            data_type = left_data_types[0] + right_data_types[0]
        self.data_type = data_type
        self.style = 'C'  # there should be a better way to do this

    def as_type(self, pdt):
        left = [x.as_type(pdt) for x in self.left]
        right = [x.as_type(pdt) for x in self.right]
        return BinaryOp(self.op, left, right)

    def _str_repr_(self, lstr, rstr):
        if '__call__' == self.op:
            return f'{lstr}({rstr})'
        if '__struct_access__' == self.op:
            return f'{lstr}.{rstr}' if 'C' == self.style else f'getattr({lstr}, "{rstr}")'
        if '__pointer_access__' == self.op:
            return f'{lstr}->{rstr}' if 'C' == self.style else f'getattr({lstr}, "{rstr}")'
        if '__getitem__' == self.op:
            return f'{lstr}[{rstr}]'
        if '__pow__' == self.op:
            return f'{lstr}^{rstr}' if 'C' == self.style else f'{lstr}**{rstr}'
        if '__lt__' == self.op:
            return f'{lstr}<{rstr}'
        if '__gt__' == self.op:
            return f'{lstr}>{rstr}'
        if '__le__' == self.op:
            return f'{lstr}<={rstr}'
        if '__ge__' == self.op:
            return f'{lstr}>={rstr}'
        if '__eq__' == self.op:
            return f'{lstr}=={rstr}'
        if '__or__' == self.op:
            return f'{lstr} || {rstr}' if 'C' == self.style else f'{lstr} or {rstr}'
        if '__and__' == self.op:
            return f'{lstr} && {rstr}' if 'C' == self.style else f'{lstr} and {rstr}'
        if any(x in self.op for x in '+-'):
            return f'({lstr} {self.op} {rstr})'
        if self.op == '//' and 'C' == self.style:
            # Verify that the operands are integers before reducing to a single slash?
            return f'{lstr} / {rstr}'
        if any(x in self.op for x in '*/'):
            return f'{lstr} {self.op} {rstr}'
        return f'{self.op}({lstr}, {rstr})'

    def __str__(self):
        return self._str_repr_(_comb(str, self.left), _comb(str, self.right))

    def __format__(self, format_spec):
        return self._str_repr_(_fmt_comb(format_spec, self.left), _fmt_comb(format_spec, self.right))

    def __repr__(self):
        return self._str_repr_(_comb(repr, self.left), _comb(repr, self.right))

    def __eq__(self, other):
        if not isinstance(other, BinaryOp):
            return False
        if self.op != other.op:
            return False
        if self.op == '*' or self.op == '+':
            # Have we implemented any other commutative operations?
            if self.left == other.right and self.right == other.left:
                return True
        return self.left == other.left and self.right == other.right

    def __round__(self, n=None):
        return self if self.data_type.is_int else UnaryOp('round', self)

    @property
    def is_scalar(self):
        return len(self.left) == 1 and len(self.right) == 1 and self.left[0].is_scalar and self.right[0].is_scalar

    @property
    def is_vector(self):
        return len(self.left) == 1 and len(self.right) == 1 and self.left[0].is_vector or self.right[0].is_vector

    @property
    def vector_known(self):
        return len(self.left) == 1 and len(self.right) == 1 and self.left[0].vector_known and self.right[0].vector_known
    #
    # def __len__(self):
    #     return max(len(self.left), len(self.right))

    def simplify(self):
        left = [x.simplify() for x in self.left]
        right = [x.simplify() for x in self.right]
        if len(left) == 1 and ((left[0].is_zero and self.op == '+') or (left[0].is_value(1) and self.op == '*')):
            return Expr(right)
        if len(right) == 1 and (
                (right[0].is_zero and any(x == self.op for x in '+-')) or
                (right[0].is_value(1) and any(x == self.op for x in '*/'))
        ):
            return Expr(left)
        if len(left) == 1 and len(right) == 1 and left[0].is_constant and right[0].is_constant\
                and self.op in ('+', '-', '*', '/'):
            if self.op == '+':
                return left[0] + right[0]
            if self.op == '-':
                return left[0] - right[0]
            if self.op == '*':
                return left[0] * right[0]
            if self.op == '/':
                return left[0] / right[0]
        # punt!
        return BinaryOp(self.op, left, right)

    def evaluate(self, known: dict):
        left, right = [[x.evaluate(known) for x in y] for y in (self.left, self.right)]
        return BinaryOp(self.op, left, right).simplify()

    def depends_on(self, name: str):
        return any(any(x.depends_on(name) for x in y) for y in (self.left, self.right))

    def copy(self):
        return BinaryOp(self.op, self.left.copy(), self.right.copy())

    def __contains__(self, value):
        # left and right are lists, so we need to check each element to avoid using __eq__
        return any(value in x for x in self.left) or any(value in x for x in self.right)

    def verify_parameters(self, instrument_parameter_names: list[str]):
        for lr in (self.left, self.right):
            for x in lr:
                x.verify_parameters(instrument_parameter_names)


class UnaryOp(Op):
    def __init__(self, op, value):
        super().__init__()
        if isinstance(value, Expr):
            value = value.expr
        if not isinstance(value, list):
            value = [value]
        self.op = op
        self.value = value
        data_types = list(dict.fromkeys(x.data_type for x in value))
        if len(data_types) != 1:
            raise RuntimeError('Multiple data types in one value not supported')
        self.data_type = data_types[0]
        self.style = 'C'  # there should be a better way to do this

    def as_type(self, pdt):
        value = [x.as_type(pdt) for x in self.value]
        return UnaryOp(self.op, value)

    def _str_repr_(self, vstr):
        if '__group__' == self.op:
            return f'({vstr})'
        if '__not__' == self.op:
            return f'!{vstr}' if 'C' == self.style else f'not {vstr}'
        if any(x in self.op for x in '+-'):
            return f'{self.op}{vstr}'
        return f'{self.op}({vstr})'

    def __str__(self):
        return self._str_repr_(_comb(str, self.value))

    def __format__(self, format_spec):
        return self._str_repr_(_fmt_comb(format_spec, self.value))

    def __repr__(self):
        return self._str_repr_(_comb(repr, self.value))

    def __neg__(self):
        if self.op == '-':
            # Avoid returning a list unless we need to
            return self.value if len(self.value) != 1 else self.value[0]
        return UnaryOp('-', self)

    def __abs__(self):
        if self.op == 'abs':
            # abs(abs(x)) is abs(x)
            return self
        return UnaryOp('abs', self)

    def __eq__(self, other):
        if not isinstance(other, UnaryOp):
            return False
        return self.op == other.op and self.value == other.value

    def __round__(self, n=None):
        return self if self.data_type.is_int else UnaryOp('round', self)

    @property
    def is_scalar(self):
        return len(self.value) == 1 and self.value[0].is_scalar

    @property
    def is_vector(self):
        return len(self.value) == 1 and self.value[0].is_vector

    @property
    def vector_known(self):
        return len(self.value) == 1 and self.value[0].vector_known

    # def __len__(self):
    #     return len(self.value)

    def __gt__(self, other):
        log.debug(f'{self} > {other} has been called (but probably should not have been!)')
        return False

    def simplify(self):
        value = [v.simplify() for v in self.value]
        if self.op == '__group__' and len(value) == 1 and isinstance(value[0], Value):
            return value[0]  # Expr(value)
        return UnaryOp(self.op, value)

    def evaluate(self, known: dict):
        value = [x.evaluate(known) for x in self.value]
        return UnaryOp(self.op, value).simplify()

    def depends_on(self, name: str):
        return any(x.depends_on(name) for x in self.value)

    def copy(self):
        return UnaryOp(self.op, self.value.copy())

    def __contains__(self, value):
        # value is a list
        return any(value in x for x in self.value)

    def verify_parameters(self, instrument_parameter_names: list[str]):
        for x in self.value:
            x.verify_parameters(instrument_parameter_names)


class Value:
    def __init__(self, value, data_type=None, object_type=None, shape_type=None):
        self._value = value
        if data_type is None or not isinstance(data_type, DataType):
            data_type = DataType.undefined
        if object_type is None or not isinstance(object_type, ObjectType):
            object_type = ObjectType.identifier if data_type != DataType.str and isinstance(value, str) else ObjectType.value
        if shape_type is None or not isinstance(shape_type, ShapeType):
            shape_type = ShapeType.vector if not isinstance(value, str) and hasattr(value, '__len__') else ShapeType.scalar
        self._object = object_type
        self._data = data_type
        self._shape = shape_type

    def __int__(self):
        if self.data_type != DataType.int:
            raise RuntimeError('Non-integer data type Value; round first')
        return self._value

    @property
    def value(self):
        return self._value

    @property
    def object_type(self):
        return self._object

    @property
    def data_type(self):
        return self._data

    @property
    def shape_type(self):
        return self._shape

    @shape_type.setter
    def shape_type(self, st):
        if not isinstance(st, ShapeType):
            raise RuntimeError('Non ShapeType value set for shape_type')
        if st.is_vector:
            if self.is_str:
                raise RuntimeError('No support for vectors of strings, e.g. char**')
            if not self.is_id or not hasattr(self.value, '__len__'):
                raise RuntimeError('Can not make a scalar value have vector type unless it is an identifier')
            self._shape = st
        else:
            if not (self.is_str or self.is_id) and hasattr(self.value, '__len__'):
                raise RuntimeError('Can not make vector value have scalar type')
            self._shape = st

    @value.setter
    def value(self, value):
        log.debug(f'Updating Value from {self._value} to {value}')
        self._object = self.data_type != DataType.str and isinstance(value, str)
        self._value = value

    @data_type.setter
    def data_type(self, dt):
        self._data = dt

    @property
    def has_value(self):
        return self.value is not None

    @property
    def vector_known(self):
        return self.is_vector and self.has_value and not isinstance(self.value, str)

    def special_str(self, prefix=None):
        if prefix is None:
            prefix = "_instrument_var._parameters."
        return f'{prefix}{self.value}' if self.is_parameter else f'{self.value}'

    def _str_repr_(self):
        return str(self.value)

    def __str__(self):
        return self._str_repr_()

    def __format__(self, format_spec):
        """Abuse string format specifications to prepend the _instrument_var._parameters. prefix to parameters"""
        if format_spec == 'p':
            return self.special_str()
        elif format_spec.startswith('prefix:'):
            return self.special_str(format_spec[7:])
        return self._str_repr_()

    def __repr__(self):
        return f'{self.shape_type} {self.data_type} {self._str_repr_()}'

    def __hash__(self):
        return hash(str(self))

    def compatible(self, other, id_ok=False):
        if isinstance(other, Expr) and other.is_singular:
            other = other.expr[0]
        if isinstance(other, (UnaryOp, BinaryOp)):
            return id_ok
        value = other if isinstance(other, Value) else Value.best(other)
        return (id_ok and value.is_str) or (self.data_type.compatible(value.data_type) and self.shape_type.compatible(value.shape_type))

    @classmethod
    def float(cls, value):
        try:
            v = float(value) if value is not None else None
        except ValueError:
            v = value
        return cls(v, DataType.float, ObjectType.value, ShapeType.scalar)

    @classmethod
    def int(cls, value):
        try:
            v = int(value) if value is not None else None
        except ValueError:
            v = value
        return cls(v, DataType.int, ObjectType.value, ShapeType.scalar)

    @classmethod
    def str(cls, value):
        return cls(value, DataType.str, ObjectType.value, ShapeType.scalar)

    @classmethod
    def id(cls, value):
        return cls(value, DataType.undefined, ObjectType.identifier, ShapeType.scalar)

    @classmethod
    def array(cls, value, dt=None):
        return cls(value, data_type=dt, object_type=None, shape_type=ShapeType.vector)

    @classmethod
    def function(cls, value, dt=None):
        return cls(value, object_type=ObjectType.function)

    @classmethod
    def best(cls, value):
        if isinstance(value, str) and value[0] == '"' and value[-1] == '"':
            return cls(value, DataType.str)
        elif isinstance(value, str):
            # Any string value which is not wrapped in double quotes must(?) be an identifier
            return cls(value, DataType.undefined, object_type=ObjectType.identifier, shape_type=ShapeType.unknown)
        if isinstance(value, int) or (isinstance(value, float) and value.is_integer()):
            return cls(value, DataType.int)
        return cls(value, DataType.float)

    @property
    def is_id(self):
        # FIXME 2023-10-16 Should instrument parameters not also be identifiers?
        return self.object_type == ObjectType.identifier or self.object_type == ObjectType.parameter

    @property
    def is_constant(self):
        return not self.is_id

    @property
    def is_parameter(self):
        return self.object_type.is_parameter

    @property
    def is_str(self):
        return self.object_type == ObjectType.value and self.data_type.is_str

    @property
    def is_float(self):
        return self.data_type == DataType.float

    @property
    def is_int(self):
        return self.data_type == DataType.int

    @property
    def is_op(self):
        return False

    @property
    def is_zero(self):
        if self.is_id:
            return False
        if self.is_str:
            # This is not great, but captures a case where, e.g., -1 is interpreted as an empty string minus 1
            return len(self.value.strip('"')) == 0
        return self.value == 0

    def is_value(self, v):
        return not self.is_id and (v.is_value(self.value) if hasattr(v, 'is_value') else self.value == v)

    @property
    def is_scalar(self):
        return self.shape_type.is_scalar

    @property
    def is_vector(self):
        return self.shape_type.is_vector

    # def __len__(self):
    #     return len(self.value) if self.is_vector else 1

    @property
    def vector_len(self):
        return len(self.value) if self.is_vector else 1

    def as_type(self, dt):
        return Value(self.value, data_type=dt, object_type=self.object_type, shape_type=self.shape_type)

    def __add__(self, other):
        other = other if isinstance(other, (Value, Op)) else Value.best(other)
        if self.is_zero:
            return other
        if other.is_zero:
            return self
        if other.is_op and isinstance(other, UnaryOp) and other.op == '-' and len(other.value) == 1:
            return self - other.value[0]
        if self.is_id and (isinstance(other, Value) and not isinstance(other.value, str) and other < 0):
            return self - (-other.value)
        if other.is_op or self.is_id or other.is_id or not self.is_constant or not other.is_constant:
            return BinaryOp('+', self, other)
        pdt = self.data_type + other.data_type
        return BinaryOp('+', self, other) if pdt.is_str else Value(self.value + other.value, pdt)

    def __sub__(self, other):
        other = other if isinstance(other, (Value, Op)) else Value.best(other)
        if self.is_zero:
            return -other
        if other.is_zero:
            return self
        if other.is_op and isinstance(other, UnaryOp) and other.op == '-' and len(other.value) == 1:
            return self + other.value[0]
        if self.is_id and (isinstance(other, Value) and not isinstance(other.value, str) and other < 0):
            return self + (-other.value)
        if other.is_op or self.is_id or other.is_id or not self.is_constant or not other.is_constant:
            return BinaryOp('-', self, other)
        pdt = self.data_type - other.data_type
        return BinaryOp('-', self, other) if pdt.is_str else Value(self.value - other.value, pdt)

    def __mul__(self, other):
        other = other if isinstance(other, (Value, Op)) else Value.best(other)
        pdt = self.data_type * other.data_type
        if self.is_zero or other.is_zero:
            return Value(0, DataType.int if pdt.is_str else pdt)
        if self.is_value(1):
            return other.as_type(pdt)
        if self.is_value(-1):
            return (-other).as_type(pdt)
        if other.is_value(1):
            return (self).as_type(pdt)
        if other.is_value(-1):
            return (-self).as_type(pdt)
        if other.is_op or self.is_id or other.is_id:
            return BinaryOp('*', self, other)
        return BinaryOp('*', self, other) if pdt.is_str else Value(self.value * other.value, pdt)

    def __truediv__(self, other):
        other = other if isinstance(other, (Value, Op)) else Value.best(other)
        pdt = self.data_type / other.data_type
        if self.is_zero:
            return Value(0, DataType.int if pdt.is_str else pdt)
        if other.is_value(1):
            return (self).as_type(pdt)
        if other.is_value(-1):
            return (-self).as_type(pdt)
        if other.is_zero:
            raise RuntimeError('Division by zero!')
        if other.is_op or self.is_id or other.is_id:
            return BinaryOp('/', self, other)
        return BinaryOp('/', self, other) if pdt.is_str else Value(self.value / other.value, pdt)

    def __floordiv__(self, other):
        other = other if isinstance(other, (Value, Op)) else Value.best(other)
        pdt = self.data_type // other.data_type
        if self.is_zero:
            return Value(0, DataType.int)
        if other.is_value(1):
            return Value.int(round(self.value))
        if other.is_value(-1):
            return -Value.int(round(self.value))
        if other.is_zero:
            raise RuntimeError('Division by zero!')
        if other.is_op or self.is_id or other.is_id:
            return BinaryOp('//', self, other)
        return BinaryOp('//', self, other) if pdt.is_str else Value.int(self.value // other.value)

    def __neg__(self):
        return UnaryOp('-', self) if self.is_id or self.data_type.is_str else Value(-self.value, self.data_type)

    def __pos__(self):
        return Value(self.value, self.data_type)

    def __abs__(self):
        return UnaryOp('abs', self) if self.is_id or self.data_type.is_str else Value(abs(self.value), self.data_type)

    def __round__(self, n=None):
        return UnaryOp('round', self) if self.is_id or self.data_type.is_str \
            else Value(round(self.value, n), self.data_type)

    def __eq__(self, other):
        other = other if isinstance(other, (Value, Op)) else Value.best(other)
        if other.is_op:
            return False
        return self.value == other.value

    def __lt__(self, other):
        other = other if isinstance(other, (Value, Op)) else Value.best(other)
        if self.is_id or other.is_op or other.is_id:
            return BinaryOp('__lt__', self, other)
        return self.value < other.value

    def __gt__(self, other):
        other = other if isinstance(other, (Value, Op)) else Value.best(other)
        if self.is_id or other.is_op or other.is_id:
            return BinaryOp('__gt__', self, other)
        return self.value > other.value

    def __le__(self, other):
        other = other if isinstance(other, (Value, Op)) else Value.best(other)
        if self.is_id or other.is_op or other.is_id:
            return BinaryOp('__le__', self, other)
        return self.value <= other.value

    def __ge__(self, other):
        other = other if isinstance(other, (Value, Op)) else Value.best(other)
        if self.is_id or other.is_op or other.is_id:
            return BinaryOp('__ge__', self, other)
        return self.value >= other.value

    def __pow__(self, power):
        if not isinstance(power, (Value, Op)):
            power = Value.best(power)
        if self.is_zero or self.is_value(1):
            return self
        if power.is_zero:
            return Value(1, data_type=self.data_type)
        return BinaryOp('__pow__', self, power)

    @property
    def mccode_c_type(self):
        return self.data_type.mccode_c_type + self.shape_type.mccode_c_type

    @property
    def mccode_c_type_name(self):
        if self.data_type == DataType.float and self.shape_type == ShapeType.scalar:
            return "instr_type_double"
        if self.data_type == DataType.int and self.shape_type == ShapeType.scalar:
            return "instr_type_int"
        if self.data_type == DataType.str and self.shape_type == ShapeType.scalar:
            return "instr_type_string"
        if self.data_type == DataType.float and self.shape_type == ShapeType.vector:
            return "instr_type_vector"
        if self.data_type == DataType.int and self.shape_type == ShapeType.vector:
            return "instr_type_vector"
        raise RuntimeError(f"No known conversion from non-enumerated data type {self.data_type} + {self.shape_type}")

    def simplify(self):
        return self

    def evaluate(self, known: dict):
        if not self.is_constant and self.value in known:
            result = known[self.value]
            if isinstance(result, Expr) and result.is_singular:
                return result.expr[0]
            return result
        return self

    def depends_on(self, name: str):
        return not self.is_constant and self.value == name

    def copy(self):
        return Value(self.value, self.data_type, self.object_type, self.shape_type)

    def __contains__(self, value):
        if self.is_id and isinstance(value, (str, Value)):
            return self.value == value
        if self.is_vector:
            return value in self.value
        if self.is_str and isinstance(value, str) and (value[0] != '"' or value[-1] != '"'):
            # string Values are always wrapped in double quotes
            return self.value.strip('"') == value.strip('"')
        return self.value == value

    def verify_parameters(self, instrument_parameter_names: list[str]):
        if self.is_id and self.value in instrument_parameter_names:
            self._object = ObjectType.parameter



class Expr:
    def __init__(self, expr: Union[Value, UnaryOp, BinaryOp, list[Union[Value, UnaryOp, BinaryOp]]]):
        self.expr = expr if isinstance(expr, list) else [expr]

    def __str__(self):
        return ','.join(str(x) for x in self.expr)

    def __format__(self, format_spec):
        """Abuse string formatting to append the _instrument_var.parameters prefix to parameter names"""
        return ','.join(format(x, format_spec) for x in self.expr)

    def __repr__(self):
        return ','.join(repr(x) for x in self.expr)

    def __hash__(self):
        return hash(str(self))

    def __contains__(self, value):
        return any(value in x for x in self.expr)

    @staticmethod
    def parse(s: str):
        from antlr4 import CommonTokenStream, InputStream
        from ..grammar import McInstrParser, McInstrLexer
        from ..instr import InstrVisitor
        parser = McInstrParser(CommonTokenStream(McInstrLexer(InputStream(s))))
        visitor = InstrVisitor(None, None)
        return visitor.getExpr(parser.expr())

    @classmethod
    def float(cls, value):
        if isinstance(value, cls):
            # Make sure the held expression *thinks* it's a float:
            for expr in value.expr:
                if expr.data_type != DataType.float:
                    expr.data_type = DataType.float
            return value
        return cls(Value.float(value))

    @classmethod
    def int(cls, value):
        if isinstance(value, cls):
            # Make sure the held expression *thinks* it's a float:
            for expr in value.expr:
                if expr.data_type != DataType.int:
                    expr.data_type = DataType.int
            return value
        return cls(Value.int(value))

    @classmethod
    def str(cls, value):
        if isinstance(value, cls):
            # Make sure the held expression *thinks* it's a float:
            for expr in value.expr:
                if expr.data_type != DataType.str:
                    expr.data_type = DataType.str
            return value
        return cls(Value.str(value))

    @classmethod
    def id(cls, value):
        if isinstance(value, cls):
            return value
        return cls(Value.id(value))

    @classmethod
    def array(cls, value):
        return cls(Value.array(value))

    @classmethod
    def function(cls, value):
        return cls(Value.function(value))

    @classmethod
    def best(cls, value):
        return cls(Value.best(value))

    @property
    def is_singular(self):
        return len(self.expr) == 1

    @property
    def is_op(self):
        return len(self.expr) == 1 and self.expr[0].is_op

    @property
    def is_zero(self):
        return len(self.expr) == 1 and self.expr[0].is_zero

    @property
    def is_id(self):
        return len(self.expr) == 1 and self.expr[0].is_id

    @property
    def is_parameter(self):
        return len(self.expr) == 1 and self.expr[0].is_parameter

    @property
    def is_str(self):
        return len(self.expr) == 1 and self.expr[0].is_str

    @property
    def is_scalar(self):
        return len(self.expr) == 1 and self.expr[0].is_scalar

    def is_value(self, value):
        return len(self.expr) == 1 and self.expr[0].is_value(value)

    @property
    def is_vector(self):
        return len(self.expr) == 1 and self.expr[0].is_vector
    #
    # def __len__(self):
    #     return len(self.expr)
    @property
    def vector_len(self):
        if len(self.expr) != 1:
            raise RuntimeError('No vector_len for array Expr objects')
        return self.expr[0].vector_len

    @property
    def is_constant(self):
        return len(self.expr) == 1 and isinstance(self.expr[0], Value) and not self.expr[0].is_id

    @property
    def has_value(self):
        return self.is_constant and self.expr[0].has_value

    @property
    def vector_known(self):
        return self.is_constant and self.expr[0].vector_known

    @property
    def value(self):
        if not self.is_constant:
            raise NotImplementedError("No conversion from expressions to constants ... yet")
        return self.expr[0].value

    @property
    def first(self):
        return self.expr[0]

    @property
    def last(self):
        return self.expr[-1]

    def compatible(self, other, id_ok=False):
        other_expr = other.expr[0] if (isinstance(other, Expr) and len(other.expr) == 1) else other
        return len(self.expr) == 1 and self.expr[0].compatible(other_expr, id_ok)

    def _prep_numeric_operation(self, msg: str, other):
        if len(self.expr) != 1:
            raise RuntimeError(f'Can not {msg} array Expr')
        return other.expr[0] if (isinstance(other, Expr) and len(other.expr) == 1) else other

    def _prep_rev_numeric_operation(self, msg: str, other):
        r = self._prep_numeric_operation(msg, other)
        if not isinstance(r, (Expr, TrinaryOp, BinaryOp, UnaryOp, Value)):
            r = Value.best(r)
        return r

    def __add__(self, other):
        return Expr(self.expr[0] + self._prep_numeric_operation('add to', other))

    def __sub__(self, other):
        return Expr(self.expr[0] - self._prep_numeric_operation('subtract', other))

    def __mul__(self, other):
        return Expr(self.expr[0] * self._prep_numeric_operation('multiply', other))

    def __truediv__(self, other):
        return Expr(self.expr[0] / self._prep_numeric_operation('divide', other))

    def __floordiv__(self, other):
        return Expr(self.expr[0] // self._prep_numeric_operation('divide', other))

    def __pow__(self, other):
        return Expr(self.expr[0] ** self._prep_numeric_operation('raise', other))

    def __radd__(self, other):
        return Expr(self._prep_rev_numeric_operation('add to', other) + self.expr[0])

    def __rsub__(self, other):
        return Expr(self._prep_rev_numeric_operation('subtract', other) - self.expr[0])

    def __rmul__(self, other):
        return Expr(self._prep_rev_numeric_operation('multiply', other) * self.expr[0])

    def __rtruediv__(self, other):
        return Expr(self._prep_rev_numeric_operation('divide', other) / self.expr[0])

    def __rfloordiv__(self, other):
        return Expr(self._prep_rev_numeric_operation('divide', other) // self.expr[0])

    def __rpow__(self, other):
        return Expr(self._prep_rev_numeric_operation('raise', other) ** self.expr[0])

    def __neg__(self):
        return Expr([-x for x in self.expr])

    def __pos__(self):
        return self

    def __abs__(self):
        return Expr([abs(x) for x in self.expr])

    def __round__(self, n=None):
        return Expr([round(x, n) for x in self.expr])

    def __eq__(self, other):
        if isinstance(other, Expr):
            if len(other.expr) != len(self.expr):
                return False
            for o_expr, s_expr in zip(other.expr, self.expr):
                if o_expr != s_expr:
                    return False
            return True
        return len(self.expr) == 1 and self.expr[0] == other

    def __lt__(self, other):
        if isinstance(other, Expr):
            if len(other.expr) != len(self.expr):
                raise RuntimeError('Can not compare unequal-sized-array Expr objects')
            for o_expr, s_expr in zip(other.expr, self.expr):
                if o_expr <= s_expr:
                    return False
            return True
        return len(self.expr) == 1 and self.expr[0] < other

    def __gt__(self, other):
        if isinstance(other, Expr):
            if len(other.expr) != len(self.expr):
                raise RuntimeError('Can not compare unequal-sized-array Expr objects')
            for o_expr, s_expr in zip(other.expr, self.expr):
                if o_expr >= s_expr:
                    return False
            return True
        return len(self.expr) == 1 and self.expr[0] > other

    def __le__(self, other):
        if isinstance(other, Expr):
            if len(other.expr) != len(self.expr):
                raise RuntimeError('Can not compare unequal-sized-array Expr objects')
            for o_expr, s_expr in zip(other.expr, self.expr):
                if o_expr < s_expr:
                    return False
            return True
        return len(self.expr) == 1 and self.expr[0] <= other

    def __ge__(self, other):
        if isinstance(other, Expr):
            if len(other.expr) != len(self.expr):
                raise RuntimeError('Can not compare unequal-sized-array Expr objects')
            for o_expr, s_expr in zip(other.expr, self.expr):
                if o_expr > s_expr:
                    return False
            return True
        return len(self.expr) == 1 and self.expr[0] >= other

    def __int__(self):
        if not len(self.expr) == 1:
            raise RuntimeError('No conversion to int for array Expr objects')
        return int(self.expr[0])

    @property
    def mccode_c_type(self):
        if len(self.expr) != 1:
            raise RuntimeError('No McCode C type for array Expr objects')
        if not isinstance(self.expr[0], Value):
            log.critical(f'Why is {self.expr[0]} not a Value?')
        return self.expr[0].mccode_c_type

    @property
    def mccode_c_type_name(self):
        if len(self.expr) != 1:
            raise RuntimeError('No McCode C type name for array Expr objects')
        return self.expr[0].mccode_c_type_name

    @property
    def data_type(self):
        if len(self.expr) != 1:
            raise RuntimeError('No data type for array Expr objects')
        return self.expr[0].data_type

    @data_type.setter
    def data_type(self, dt):
        if len(self.expr) != 1:
            raise RuntimeError('No data type for array Expr objects')
        self.expr[0].data_type = dt

    @property
    def shape_type(self):
        if len(self.expr) != 1:
            raise RuntimeError('No data type for array Expr objects')
        return self.expr[0].shape_type

    @shape_type.setter
    def shape_type(self, st):
        if len(self.expr) != 1:
            raise RuntimeError('No data type for array Expr objects')
        self.expr[0].shape_type = st

    def simplify(self):
        """Perform a very basic analysis to reduce the expression complexity"""
        return Expr([x.simplify() for x in self.expr])

    def evaluate(self, known: dict):
        return Expr([x.evaluate(known) for x in self.expr]).simplify()

    def depends_on(self, name: str):
        return any(x.depends_on(name) for x in self.expr)

    def copy(self):
        return Expr([x.copy() for x in self.expr])

    def verify_parameters(self, instrument_parameter_names: list[str]):
        for x in self.expr:
            x.verify_parameters(instrument_parameter_names)


def unary_expr(func, name, v):
    ops = {'cos': 'acos', 'sin': 'asin', 'tan': 'atan'}
    if isinstance(v, Expr):
        v = v.expr
    if len(v) == 1:
        v = v[0]
    if isinstance(v, UnaryOp) and ((name in ops and v.op == ops[name]) or (v.op in ops and name == ops[v.op])):
        return Expr(v.value)
    if isinstance(v, Value) and not v.is_id:
        if v.is_str or isinstance(v.value, str):
            raise RuntimeError(f'How is a _string_ valued parameter, {v} not an identifier?')
        return Expr(Value.best(func(v.value)))
    return Expr(UnaryOp(name, v))


def binary_expr(func, name, left, right):
    ops = {'atan2': ('sin', 'cos')}
    if isinstance(left, Expr):
        left = left.expr
    if isinstance(right, Expr):
        right = right.expr
    if len(left) == 1:
        left = left[0]
    if len(right) == 1:
        right = right[0]
    if isinstance(left, UnaryOp) and isinstance(right, UnaryOp) and name in ops:
        left_func, right_func = ops[name]
        if left.op == left_func and right.op == right_func and left.value == right.value:
            return Expr(left.value)
    if isinstance(left, Value) and isinstance(right, Value) and not left.is_id and not right.is_id:
        return Expr(Value.best(func(left.value, right.value)))
    return Expr(BinaryOp(name, left, right))



# Generated from /home/g/Code/mccode_antlr-antlr/mccode_antlr/grammar/McInstr.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .McInstrParser import McInstrParser
else:
    from McInstrParser import McInstrParser

# This class defines a complete listener for a parse tree produced by McInstrParser.
class McInstrListener(ParseTreeListener):

    # Enter a parse tree produced by McInstrParser#prog.
    def enterProg(self, ctx:McInstrParser.ProgContext):
        pass

    # Exit a parse tree produced by McInstrParser#prog.
    def exitProg(self, ctx:McInstrParser.ProgContext):
        pass


    # Enter a parse tree produced by McInstrParser#instrument_definition.
    def enterInstrument_definition(self, ctx:McInstrParser.Instrument_definitionContext):
        pass

    # Exit a parse tree produced by McInstrParser#instrument_definition.
    def exitInstrument_definition(self, ctx:McInstrParser.Instrument_definitionContext):
        pass


    # Enter a parse tree produced by McInstrParser#instrument_parameters.
    def enterInstrument_parameters(self, ctx:McInstrParser.Instrument_parametersContext):
        pass

    # Exit a parse tree produced by McInstrParser#instrument_parameters.
    def exitInstrument_parameters(self, ctx:McInstrParser.Instrument_parametersContext):
        pass


    # Enter a parse tree produced by McInstrParser#InstrumentParameterDouble.
    def enterInstrumentParameterDouble(self, ctx:McInstrParser.InstrumentParameterDoubleContext):
        pass

    # Exit a parse tree produced by McInstrParser#InstrumentParameterDouble.
    def exitInstrumentParameterDouble(self, ctx:McInstrParser.InstrumentParameterDoubleContext):
        pass


    # Enter a parse tree produced by McInstrParser#InstrumentParameterInteger.
    def enterInstrumentParameterInteger(self, ctx:McInstrParser.InstrumentParameterIntegerContext):
        pass

    # Exit a parse tree produced by McInstrParser#InstrumentParameterInteger.
    def exitInstrumentParameterInteger(self, ctx:McInstrParser.InstrumentParameterIntegerContext):
        pass


    # Enter a parse tree produced by McInstrParser#InstrumentParameterString.
    def enterInstrumentParameterString(self, ctx:McInstrParser.InstrumentParameterStringContext):
        pass

    # Exit a parse tree produced by McInstrParser#InstrumentParameterString.
    def exitInstrumentParameterString(self, ctx:McInstrParser.InstrumentParameterStringContext):
        pass


    # Enter a parse tree produced by McInstrParser#instrument_parameter_unit.
    def enterInstrument_parameter_unit(self, ctx:McInstrParser.Instrument_parameter_unitContext):
        pass

    # Exit a parse tree produced by McInstrParser#instrument_parameter_unit.
    def exitInstrument_parameter_unit(self, ctx:McInstrParser.Instrument_parameter_unitContext):
        pass


    # Enter a parse tree produced by McInstrParser#instrument_trace.
    def enterInstrument_trace(self, ctx:McInstrParser.Instrument_traceContext):
        pass

    # Exit a parse tree produced by McInstrParser#instrument_trace.
    def exitInstrument_trace(self, ctx:McInstrParser.Instrument_traceContext):
        pass


    # Enter a parse tree produced by McInstrParser#instrument_metadata.
    def enterInstrument_metadata(self, ctx:McInstrParser.Instrument_metadataContext):
        pass

    # Exit a parse tree produced by McInstrParser#instrument_metadata.
    def exitInstrument_metadata(self, ctx:McInstrParser.Instrument_metadataContext):
        pass


    # Enter a parse tree produced by McInstrParser#instrument_trace_include.
    def enterInstrument_trace_include(self, ctx:McInstrParser.Instrument_trace_includeContext):
        pass

    # Exit a parse tree produced by McInstrParser#instrument_trace_include.
    def exitInstrument_trace_include(self, ctx:McInstrParser.Instrument_trace_includeContext):
        pass


    # Enter a parse tree produced by McInstrParser#component_instance.
    def enterComponent_instance(self, ctx:McInstrParser.Component_instanceContext):
        pass

    # Exit a parse tree produced by McInstrParser#component_instance.
    def exitComponent_instance(self, ctx:McInstrParser.Component_instanceContext):
        pass


    # Enter a parse tree produced by McInstrParser#InstanceNameCopyIdentifier.
    def enterInstanceNameCopyIdentifier(self, ctx:McInstrParser.InstanceNameCopyIdentifierContext):
        pass

    # Exit a parse tree produced by McInstrParser#InstanceNameCopyIdentifier.
    def exitInstanceNameCopyIdentifier(self, ctx:McInstrParser.InstanceNameCopyIdentifierContext):
        pass


    # Enter a parse tree produced by McInstrParser#InstanceNameCopy.
    def enterInstanceNameCopy(self, ctx:McInstrParser.InstanceNameCopyContext):
        pass

    # Exit a parse tree produced by McInstrParser#InstanceNameCopy.
    def exitInstanceNameCopy(self, ctx:McInstrParser.InstanceNameCopyContext):
        pass


    # Enter a parse tree produced by McInstrParser#InstanceNameIdentifier.
    def enterInstanceNameIdentifier(self, ctx:McInstrParser.InstanceNameIdentifierContext):
        pass

    # Exit a parse tree produced by McInstrParser#InstanceNameIdentifier.
    def exitInstanceNameIdentifier(self, ctx:McInstrParser.InstanceNameIdentifierContext):
        pass


    # Enter a parse tree produced by McInstrParser#ComponentTypeCopy.
    def enterComponentTypeCopy(self, ctx:McInstrParser.ComponentTypeCopyContext):
        pass

    # Exit a parse tree produced by McInstrParser#ComponentTypeCopy.
    def exitComponentTypeCopy(self, ctx:McInstrParser.ComponentTypeCopyContext):
        pass


    # Enter a parse tree produced by McInstrParser#ComponentTypeIdentifier.
    def enterComponentTypeIdentifier(self, ctx:McInstrParser.ComponentTypeIdentifierContext):
        pass

    # Exit a parse tree produced by McInstrParser#ComponentTypeIdentifier.
    def exitComponentTypeIdentifier(self, ctx:McInstrParser.ComponentTypeIdentifierContext):
        pass


    # Enter a parse tree produced by McInstrParser#instance_parameters.
    def enterInstance_parameters(self, ctx:McInstrParser.Instance_parametersContext):
        pass

    # Exit a parse tree produced by McInstrParser#instance_parameters.
    def exitInstance_parameters(self, ctx:McInstrParser.Instance_parametersContext):
        pass


    # Enter a parse tree produced by McInstrParser#InstanceParameterExpr.
    def enterInstanceParameterExpr(self, ctx:McInstrParser.InstanceParameterExprContext):
        pass

    # Exit a parse tree produced by McInstrParser#InstanceParameterExpr.
    def exitInstanceParameterExpr(self, ctx:McInstrParser.InstanceParameterExprContext):
        pass


    # Enter a parse tree produced by McInstrParser#InstanceParameterNull.
    def enterInstanceParameterNull(self, ctx:McInstrParser.InstanceParameterNullContext):
        pass

    # Exit a parse tree produced by McInstrParser#InstanceParameterNull.
    def exitInstanceParameterNull(self, ctx:McInstrParser.InstanceParameterNullContext):
        pass


    # Enter a parse tree produced by McInstrParser#InstanceParameterVector.
    def enterInstanceParameterVector(self, ctx:McInstrParser.InstanceParameterVectorContext):
        pass

    # Exit a parse tree produced by McInstrParser#InstanceParameterVector.
    def exitInstanceParameterVector(self, ctx:McInstrParser.InstanceParameterVectorContext):
        pass


    # Enter a parse tree produced by McInstrParser#split.
    def enterSplit(self, ctx:McInstrParser.SplitContext):
        pass

    # Exit a parse tree produced by McInstrParser#split.
    def exitSplit(self, ctx:McInstrParser.SplitContext):
        pass


    # Enter a parse tree produced by McInstrParser#when.
    def enterWhen(self, ctx:McInstrParser.WhenContext):
        pass

    # Exit a parse tree produced by McInstrParser#when.
    def exitWhen(self, ctx:McInstrParser.WhenContext):
        pass


    # Enter a parse tree produced by McInstrParser#place.
    def enterPlace(self, ctx:McInstrParser.PlaceContext):
        pass

    # Exit a parse tree produced by McInstrParser#place.
    def exitPlace(self, ctx:McInstrParser.PlaceContext):
        pass


    # Enter a parse tree produced by McInstrParser#orientation.
    def enterOrientation(self, ctx:McInstrParser.OrientationContext):
        pass

    # Exit a parse tree produced by McInstrParser#orientation.
    def exitOrientation(self, ctx:McInstrParser.OrientationContext):
        pass


    # Enter a parse tree produced by McInstrParser#groupref.
    def enterGroupref(self, ctx:McInstrParser.GrouprefContext):
        pass

    # Exit a parse tree produced by McInstrParser#groupref.
    def exitGroupref(self, ctx:McInstrParser.GrouprefContext):
        pass


    # Enter a parse tree produced by McInstrParser#jumps.
    def enterJumps(self, ctx:McInstrParser.JumpsContext):
        pass

    # Exit a parse tree produced by McInstrParser#jumps.
    def exitJumps(self, ctx:McInstrParser.JumpsContext):
        pass


    # Enter a parse tree produced by McInstrParser#jump.
    def enterJump(self, ctx:McInstrParser.JumpContext):
        pass

    # Exit a parse tree produced by McInstrParser#jump.
    def exitJump(self, ctx:McInstrParser.JumpContext):
        pass


    # Enter a parse tree produced by McInstrParser#JumpPrevious.
    def enterJumpPrevious(self, ctx:McInstrParser.JumpPreviousContext):
        pass

    # Exit a parse tree produced by McInstrParser#JumpPrevious.
    def exitJumpPrevious(self, ctx:McInstrParser.JumpPreviousContext):
        pass


    # Enter a parse tree produced by McInstrParser#JumpMyself.
    def enterJumpMyself(self, ctx:McInstrParser.JumpMyselfContext):
        pass

    # Exit a parse tree produced by McInstrParser#JumpMyself.
    def exitJumpMyself(self, ctx:McInstrParser.JumpMyselfContext):
        pass


    # Enter a parse tree produced by McInstrParser#JumpNext.
    def enterJumpNext(self, ctx:McInstrParser.JumpNextContext):
        pass

    # Exit a parse tree produced by McInstrParser#JumpNext.
    def exitJumpNext(self, ctx:McInstrParser.JumpNextContext):
        pass


    # Enter a parse tree produced by McInstrParser#JumpIdentifier.
    def enterJumpIdentifier(self, ctx:McInstrParser.JumpIdentifierContext):
        pass

    # Exit a parse tree produced by McInstrParser#JumpIdentifier.
    def exitJumpIdentifier(self, ctx:McInstrParser.JumpIdentifierContext):
        pass


    # Enter a parse tree produced by McInstrParser#extend.
    def enterExtend(self, ctx:McInstrParser.ExtendContext):
        pass

    # Exit a parse tree produced by McInstrParser#extend.
    def exitExtend(self, ctx:McInstrParser.ExtendContext):
        pass


    # Enter a parse tree produced by McInstrParser#component_ref.
    def enterComponent_ref(self, ctx:McInstrParser.Component_refContext):
        pass

    # Exit a parse tree produced by McInstrParser#component_ref.
    def exitComponent_ref(self, ctx:McInstrParser.Component_refContext):
        pass


    # Enter a parse tree produced by McInstrParser#coords.
    def enterCoords(self, ctx:McInstrParser.CoordsContext):
        pass

    # Exit a parse tree produced by McInstrParser#coords.
    def exitCoords(self, ctx:McInstrParser.CoordsContext):
        pass


    # Enter a parse tree produced by McInstrParser#reference.
    def enterReference(self, ctx:McInstrParser.ReferenceContext):
        pass

    # Exit a parse tree produced by McInstrParser#reference.
    def exitReference(self, ctx:McInstrParser.ReferenceContext):
        pass


    # Enter a parse tree produced by McInstrParser#dependency.
    def enterDependency(self, ctx:McInstrParser.DependencyContext):
        pass

    # Exit a parse tree produced by McInstrParser#dependency.
    def exitDependency(self, ctx:McInstrParser.DependencyContext):
        pass


    # Enter a parse tree produced by McInstrParser#DeclareBlock.
    def enterDeclareBlock(self, ctx:McInstrParser.DeclareBlockContext):
        pass

    # Exit a parse tree produced by McInstrParser#DeclareBlock.
    def exitDeclareBlock(self, ctx:McInstrParser.DeclareBlockContext):
        pass


    # Enter a parse tree produced by McInstrParser#DeclareBlockCopy.
    def enterDeclareBlockCopy(self, ctx:McInstrParser.DeclareBlockCopyContext):
        pass

    # Exit a parse tree produced by McInstrParser#DeclareBlockCopy.
    def exitDeclareBlockCopy(self, ctx:McInstrParser.DeclareBlockCopyContext):
        pass


    # Enter a parse tree produced by McInstrParser#uservars.
    def enterUservars(self, ctx:McInstrParser.UservarsContext):
        pass

    # Exit a parse tree produced by McInstrParser#uservars.
    def exitUservars(self, ctx:McInstrParser.UservarsContext):
        pass


    # Enter a parse tree produced by McInstrParser#InitializeBlock.
    def enterInitializeBlock(self, ctx:McInstrParser.InitializeBlockContext):
        pass

    # Exit a parse tree produced by McInstrParser#InitializeBlock.
    def exitInitializeBlock(self, ctx:McInstrParser.InitializeBlockContext):
        pass


    # Enter a parse tree produced by McInstrParser#InitializeBlockCopy.
    def enterInitializeBlockCopy(self, ctx:McInstrParser.InitializeBlockCopyContext):
        pass

    # Exit a parse tree produced by McInstrParser#InitializeBlockCopy.
    def exitInitializeBlockCopy(self, ctx:McInstrParser.InitializeBlockCopyContext):
        pass


    # Enter a parse tree produced by McInstrParser#SaveBlock.
    def enterSaveBlock(self, ctx:McInstrParser.SaveBlockContext):
        pass

    # Exit a parse tree produced by McInstrParser#SaveBlock.
    def exitSaveBlock(self, ctx:McInstrParser.SaveBlockContext):
        pass


    # Enter a parse tree produced by McInstrParser#SaveBlockCopy.
    def enterSaveBlockCopy(self, ctx:McInstrParser.SaveBlockCopyContext):
        pass

    # Exit a parse tree produced by McInstrParser#SaveBlockCopy.
    def exitSaveBlockCopy(self, ctx:McInstrParser.SaveBlockCopyContext):
        pass


    # Enter a parse tree produced by McInstrParser#FinallyBlock.
    def enterFinallyBlock(self, ctx:McInstrParser.FinallyBlockContext):
        pass

    # Exit a parse tree produced by McInstrParser#FinallyBlock.
    def exitFinallyBlock(self, ctx:McInstrParser.FinallyBlockContext):
        pass


    # Enter a parse tree produced by McInstrParser#FinallyBlockCopy.
    def enterFinallyBlockCopy(self, ctx:McInstrParser.FinallyBlockCopyContext):
        pass

    # Exit a parse tree produced by McInstrParser#FinallyBlockCopy.
    def exitFinallyBlockCopy(self, ctx:McInstrParser.FinallyBlockCopyContext):
        pass


    # Enter a parse tree produced by McInstrParser#metadata.
    def enterMetadata(self, ctx:McInstrParser.MetadataContext):
        pass

    # Exit a parse tree produced by McInstrParser#metadata.
    def exitMetadata(self, ctx:McInstrParser.MetadataContext):
        pass


    # Enter a parse tree produced by McInstrParser#category.
    def enterCategory(self, ctx:McInstrParser.CategoryContext):
        pass

    # Exit a parse tree produced by McInstrParser#category.
    def exitCategory(self, ctx:McInstrParser.CategoryContext):
        pass


    # Enter a parse tree produced by McInstrParser#initializerlist.
    def enterInitializerlist(self, ctx:McInstrParser.InitializerlistContext):
        pass

    # Exit a parse tree produced by McInstrParser#initializerlist.
    def exitInitializerlist(self, ctx:McInstrParser.InitializerlistContext):
        pass


    # Enter a parse tree produced by McInstrParser#assignment.
    def enterAssignment(self, ctx:McInstrParser.AssignmentContext):
        pass

    # Exit a parse tree produced by McInstrParser#assignment.
    def exitAssignment(self, ctx:McInstrParser.AssignmentContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionBinaryLess.
    def enterExpressionBinaryLess(self, ctx:McInstrParser.ExpressionBinaryLessContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionBinaryLess.
    def exitExpressionBinaryLess(self, ctx:McInstrParser.ExpressionBinaryLessContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionBinaryGreater.
    def enterExpressionBinaryGreater(self, ctx:McInstrParser.ExpressionBinaryGreaterContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionBinaryGreater.
    def exitExpressionBinaryGreater(self, ctx:McInstrParser.ExpressionBinaryGreaterContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionGrouping.
    def enterExpressionGrouping(self, ctx:McInstrParser.ExpressionGroupingContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionGrouping.
    def exitExpressionGrouping(self, ctx:McInstrParser.ExpressionGroupingContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionBinaryLessEqual.
    def enterExpressionBinaryLessEqual(self, ctx:McInstrParser.ExpressionBinaryLessEqualContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionBinaryLessEqual.
    def exitExpressionBinaryLessEqual(self, ctx:McInstrParser.ExpressionBinaryLessEqualContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionArrayAccess.
    def enterExpressionArrayAccess(self, ctx:McInstrParser.ExpressionArrayAccessContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionArrayAccess.
    def exitExpressionArrayAccess(self, ctx:McInstrParser.ExpressionArrayAccessContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionBinaryLogic.
    def enterExpressionBinaryLogic(self, ctx:McInstrParser.ExpressionBinaryLogicContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionBinaryLogic.
    def exitExpressionBinaryLogic(self, ctx:McInstrParser.ExpressionBinaryLogicContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionInteger.
    def enterExpressionInteger(self, ctx:McInstrParser.ExpressionIntegerContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionInteger.
    def exitExpressionInteger(self, ctx:McInstrParser.ExpressionIntegerContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionExponentiation.
    def enterExpressionExponentiation(self, ctx:McInstrParser.ExpressionExponentiationContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionExponentiation.
    def exitExpressionExponentiation(self, ctx:McInstrParser.ExpressionExponentiationContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionBinaryGreaterEqual.
    def enterExpressionBinaryGreaterEqual(self, ctx:McInstrParser.ExpressionBinaryGreaterEqualContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionBinaryGreaterEqual.
    def exitExpressionBinaryGreaterEqual(self, ctx:McInstrParser.ExpressionBinaryGreaterEqualContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionZero.
    def enterExpressionZero(self, ctx:McInstrParser.ExpressionZeroContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionZero.
    def exitExpressionZero(self, ctx:McInstrParser.ExpressionZeroContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionMyself.
    def enterExpressionMyself(self, ctx:McInstrParser.ExpressionMyselfContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionMyself.
    def exitExpressionMyself(self, ctx:McInstrParser.ExpressionMyselfContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionUnaryPM.
    def enterExpressionUnaryPM(self, ctx:McInstrParser.ExpressionUnaryPMContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionUnaryPM.
    def exitExpressionUnaryPM(self, ctx:McInstrParser.ExpressionUnaryPMContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionPrevious.
    def enterExpressionPrevious(self, ctx:McInstrParser.ExpressionPreviousContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionPrevious.
    def exitExpressionPrevious(self, ctx:McInstrParser.ExpressionPreviousContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionTrinaryLogic.
    def enterExpressionTrinaryLogic(self, ctx:McInstrParser.ExpressionTrinaryLogicContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionTrinaryLogic.
    def exitExpressionTrinaryLogic(self, ctx:McInstrParser.ExpressionTrinaryLogicContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionFloat.
    def enterExpressionFloat(self, ctx:McInstrParser.ExpressionFloatContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionFloat.
    def exitExpressionFloat(self, ctx:McInstrParser.ExpressionFloatContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionPointerAccess.
    def enterExpressionPointerAccess(self, ctx:McInstrParser.ExpressionPointerAccessContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionPointerAccess.
    def exitExpressionPointerAccess(self, ctx:McInstrParser.ExpressionPointerAccessContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionIdentifier.
    def enterExpressionIdentifier(self, ctx:McInstrParser.ExpressionIdentifierContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionIdentifier.
    def exitExpressionIdentifier(self, ctx:McInstrParser.ExpressionIdentifierContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionBinaryEqual.
    def enterExpressionBinaryEqual(self, ctx:McInstrParser.ExpressionBinaryEqualContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionBinaryEqual.
    def exitExpressionBinaryEqual(self, ctx:McInstrParser.ExpressionBinaryEqualContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionBinaryPM.
    def enterExpressionBinaryPM(self, ctx:McInstrParser.ExpressionBinaryPMContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionBinaryPM.
    def exitExpressionBinaryPM(self, ctx:McInstrParser.ExpressionBinaryPMContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionUnaryLogic.
    def enterExpressionUnaryLogic(self, ctx:McInstrParser.ExpressionUnaryLogicContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionUnaryLogic.
    def exitExpressionUnaryLogic(self, ctx:McInstrParser.ExpressionUnaryLogicContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionStructAccess.
    def enterExpressionStructAccess(self, ctx:McInstrParser.ExpressionStructAccessContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionStructAccess.
    def exitExpressionStructAccess(self, ctx:McInstrParser.ExpressionStructAccessContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionFunctionCall.
    def enterExpressionFunctionCall(self, ctx:McInstrParser.ExpressionFunctionCallContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionFunctionCall.
    def exitExpressionFunctionCall(self, ctx:McInstrParser.ExpressionFunctionCallContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionBinaryMD.
    def enterExpressionBinaryMD(self, ctx:McInstrParser.ExpressionBinaryMDContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionBinaryMD.
    def exitExpressionBinaryMD(self, ctx:McInstrParser.ExpressionBinaryMDContext):
        pass


    # Enter a parse tree produced by McInstrParser#ExpressionString.
    def enterExpressionString(self, ctx:McInstrParser.ExpressionStringContext):
        pass

    # Exit a parse tree produced by McInstrParser#ExpressionString.
    def exitExpressionString(self, ctx:McInstrParser.ExpressionStringContext):
        pass


    # Enter a parse tree produced by McInstrParser#shell.
    def enterShell(self, ctx:McInstrParser.ShellContext):
        pass

    # Exit a parse tree produced by McInstrParser#shell.
    def exitShell(self, ctx:McInstrParser.ShellContext):
        pass


    # Enter a parse tree produced by McInstrParser#SearchPath.
    def enterSearchPath(self, ctx:McInstrParser.SearchPathContext):
        pass

    # Exit a parse tree produced by McInstrParser#SearchPath.
    def exitSearchPath(self, ctx:McInstrParser.SearchPathContext):
        pass


    # Enter a parse tree produced by McInstrParser#SearchShell.
    def enterSearchShell(self, ctx:McInstrParser.SearchShellContext):
        pass

    # Exit a parse tree produced by McInstrParser#SearchShell.
    def exitSearchShell(self, ctx:McInstrParser.SearchShellContext):
        pass


    # Enter a parse tree produced by McInstrParser#unparsed_block.
    def enterUnparsed_block(self, ctx:McInstrParser.Unparsed_blockContext):
        pass

    # Exit a parse tree produced by McInstrParser#unparsed_block.
    def exitUnparsed_block(self, ctx:McInstrParser.Unparsed_blockContext):
        pass



del McInstrParser